import numpy as np


class LensPositionFit(object):
    def __init__(self, positions, noise_map):
        """A lens position fitter, which takes a set of positions (e.g. from a plane in the tracer) and computes \
        their maximum separation, such that points which tracer closer to one another have a higher likelihood.

        Parameters
        -----------
        positions : [[]]
            The (y,x) arc-second coordinates of positions which the maximum distance and likelihood is computed using.
        noise_map : ndarray | float
            The noise-value assumed when computing the likelihood.
        """
        self.positions = positions
        self.noise_map = noise_map

    @property
    def chi_squared_map(self):
        return np.square(np.divide(self.maximum_separations, self.noise_map))

    @property
    def figure_of_merit(self):
        return -0.5 * sum(self.chi_squared_map)

    def maximum_separation_within_threshold(self, threshold):
        return max(self.maximum_separations) <= threshold

    @property
    def maximum_separations(self):
        return list(
            map(
                lambda positions: self.max_separation_of_grid(grid=positions),
                self.positions,
            )
        )

    @staticmethod
    def max_separation_of_grid(grid):
        rdist_max = np.zeros((grid.shape[0]))
        for i in range(grid.shape[0]):
            xdists = np.square(np.subtract(grid[i, 0], grid[:, 0]))
            ydists = np.square(np.subtract(grid[i, 1], grid[:, 1]))
            rdist_max[i] = np.max(np.add(xdists, ydists))
        return np.max(np.sqrt(rdist_max))



from astropy import cosmology as cosmo

import autofit as af
from autolens.lens import ray_tracing
from autolens.model.galaxy import galaxy as g


class AbstractPhase(af.AbstractPhase):
    def __init__(
        self,
        phase_name,
        phase_tag=None,
        phase_folders=tuple(),
        optimizer_class=af.MultiNest,
        cosmology=cosmo.Planck15,
        auto_link_priors=False,
    ):
        """
        A phase in an lens pipeline. Uses the set non_linear optimizer to try to fit
        models and hyper_galaxies passed to it.

        Parameters
        ----------
        optimizer_class: class
            The class of a non_linear optimizer
        phase_name: str
            The name of this phase
        """

        super().__init__(
            phase_name=phase_name,
            phase_tag=phase_tag,
            phase_folders=phase_folders,
            optimizer_class=optimizer_class,
            auto_link_priors=auto_link_priors,
        )

        self.cosmology = cosmology

    @property
    def phase_folders(self):
        return self.optimizer.phase_folders

    @property
    def phase_property_collections(self):
        """
        Returns
        -------
        phase_property_collections: [PhaseProperty]
            A list of phase property collections associated with this phase. This is
            used in automated prior passing and should be overridden for any phase that
            contains its own PhasePropertys.
        """
        return []

    @property
    def path(self):
        return self.optimizer.path

    def customize_priors(self, results):
        """
        Perform any prior or constant passing. This could involve setting model
        attributes equal to priors or constants from a previous phase.

        Parameters
        ----------
        results: autofit.tools.pipeline.ResultsCollection
            The result of the previous phase
        """
        pass

    # noinspection PyAbstractClass
    class Analysis(af.Analysis):
        def __init__(self, cosmology, results=None):
            """
            An lens object

            Parameters
            ----------
            results: autofit.tools.pipeline.ResultsCollection
                The results of all previous phases
            """

            self.results = results
            self.cosmology = cosmology

        @property
        def last_results(self):
            """
            Returns
            -------
            result: AbstractPhase.Result | None
                The result from the last phase
            """
            if self.results is not None:
                return self.results.last

        def tracer_for_instance(self, instance):
            return ray_tracing.Tracer.from_galaxies(
                galaxies=instance.galaxies, cosmology=self.cosmology
            )

    def make_result(self, result, analysis):
        return self.__class__.Result(
            constant=result.constant,
            figure_of_merit=result.figure_of_merit,
            previous_variable=result.previous_variable,
            gaussian_tuples=result.gaussian_tuples,
            analysis=analysis,
            optimizer=self.optimizer,
        )

    class Result(af.Result):
        def __init__(
            self,
            constant,
            figure_of_merit,
            previous_variable,
            gaussian_tuples,
            analysis,
            optimizer,
        ):
            """
            The result of a phase
            """
            super().__init__(
                constant=constant,
                figure_of_merit=figure_of_merit,
                previous_variable=previous_variable,
                gaussian_tuples=gaussian_tuples,
            )

            self.analysis = analysis
            self.optimizer = optimizer

        @property
        def most_likely_tracer(self):
            return self.analysis.tracer_for_instance(instance=self.constant)

        @property
        def path_galaxy_tuples(self) -> [(str, g.Galaxy)]:
            """
            Tuples associating the names of galaxies with instances from the best fit
            """
            return self.constant.path_instance_tuples_for_class(cls=g.Galaxy)

    def run(self, image, results=None, mask=None):
        raise NotImplementedError()



import os

import autofit as af
from autolens.model.galaxy.plotters import galaxy_fit_plotters
from autolens.pipeline.plotters import phase_plotters, hyper_plotters


def setting(section, name):
    return af.conf.instance.visualize.get(section, name, bool)


def plot_setting(name):
    return setting("plots", name)


def figure_setting(name):
    return setting("figures", name)


class AbstractVisualizer:
    def __init__(self, image_path):
        self.image_path = image_path
        self.plot_units = af.conf.instance.visualize.get(
            "figures", "plot_units", str
        ).strip()
        self.should_plot_mask = figure_setting("plot_mask_on_images")
        self.zoom_around_mask = figure_setting("zoom_around_mask_of_images")
        self.plot_ray_tracing_all_at_end_png = plot_setting(
            "plot_ray_tracing_all_at_end_png"
        )
        self.plot_ray_tracing_all_at_end_fits = plot_setting(
            "plot_ray_tracing_all_at_end_fits"
        )


class SubPlotVisualizer(AbstractVisualizer):
    def __init__(self, image_path):
        super().__init__(image_path)
        self.subplot_path = f"{image_path}subplots/"
        try:
            os.makedirs(self.subplot_path)
        except FileExistsError:
            pass


class PhaseGalaxyVisualizer(AbstractVisualizer):
    def __init__(self, image_path):
        super().__init__(image_path)
        self.plot_galaxy_fit_all_at_end_png = plot_setting(
            "plot_galaxy_fit_all_at_end_png"
        )
        self.plot_galaxy_fit_all_at_end_fits = plot_setting(
            "plot_galaxy_fit_all_at_end_fits"
        )
        self.plot_galaxy_fit_as_subplot = plot_setting("plot_galaxy_fit_as_subplot")
        self.plot_galaxy_fit_image = plot_setting("plot_galaxy_fit_image")
        self.plot_galaxy_fit_noise_map = plot_setting("plot_galaxy_fit_noise_map")
        self.plot_galaxy_fit_model_image = plot_setting("plot_galaxy_fit_model_image")
        self.plot_galaxy_fit_residual_map = plot_setting("plot_galaxy_fit_residual_map")
        self.plot_galaxy_fit_chi_squared_map = plot_setting(
            "plot_galaxy_fit_chi_squared_map"
        )

    def plot_galaxy_fit_subplot(self, fit, path_suffix=""):
        if self.plot_galaxy_fit_as_subplot:
            galaxy_fit_plotters.plot_fit_subplot(
                fit=fit,
                should_plot_mask=self.should_plot_mask,
                zoom_around_mask=self.zoom_around_mask,
                units=self.plot_units,
                output_path=f"{self.image_path}/{path_suffix}",
                output_format="png",
            )

    def plot_fit_individuals(
        self, fit, plot_all=False, image_format="png", path_suffix=""
    ):
        if plot_all:
            should_plot_image = True
            should_plot_noise_map = True
            should_plot_model_image = True
            should_plot_residual_map = True
            should_plot_chi_squared_map = True
        else:
            should_plot_image = self.plot_galaxy_fit_image
            should_plot_noise_map = self.plot_galaxy_fit_noise_map
            should_plot_model_image = self.plot_galaxy_fit_model_image
            should_plot_residual_map = self.plot_galaxy_fit_residual_map
            should_plot_chi_squared_map = self.plot_galaxy_fit_chi_squared_map
        galaxy_fit_plotters.plot_fit_individuals(
            fit=fit,
            should_plot_mask=self.should_plot_mask,
            zoom_around_mask=self.zoom_around_mask,
            should_plot_image=should_plot_image,
            should_plot_noise_map=should_plot_noise_map,
            should_plot_model_image=should_plot_model_image,
            should_plot_residual_map=should_plot_residual_map,
            should_plot_chi_squared_map=should_plot_chi_squared_map,
            units=self.plot_units,
            output_path=f"{self.image_path}/{path_suffix}",
            output_format=image_format,
        )


class PhaseImagingVisualizer(SubPlotVisualizer):
    def __init__(self, lens_imaging_data, image_path):
        super().__init__(image_path)
        self.lens_imaging_data = lens_imaging_data

        self.should_plot_image_plane_pix = figure_setting(
            "plot_image_plane_adaptive_pixelization_grid"
        )
        self.plot_data_as_subplot = plot_setting("plot_data_as_subplot")
        self.plot_data_image = plot_setting("plot_data_image")
        self.plot_data_noise_map = plot_setting("plot_data_noise_map")
        self.plot_data_psf = plot_setting("plot_data_psf")

        self.plot_data_signal_to_noise_map = plot_setting(
            "plot_data_signal_to_noise_map"
        )
        self.plot_data_absolute_signal_to_noise_map = plot_setting(
            "plot_data_absolute_signal_to_noise_map"
        )
        self.plot_data_potential_chi_squared_map = plot_setting(
            "plot_data_potential_chi_squared_map"
        )
        self.plot_lens_fit_all_at_end_png = plot_setting("plot_lens_fit_all_at_end_png")
        self.plot_lens_fit_all_at_end_fits = plot_setting(
            "plot_lens_fit_all_at_end_fits"
        )
        self.plot_lens_fit_as_subplot = plot_setting("plot_lens_fit_as_subplot")
        self.plot_lens_fit_of_planes_as_subplot = plot_setting(
            "plot_lens_fit_of_planes_as_subplot"
        )
        self.plot_lens_fit_inversion_as_subplot = plot_setting(
            "plot_lens_fit_inversion_as_subplot"
        )
        self.plot_lens_fit_image = plot_setting("plot_lens_fit_image")
        self.plot_lens_fit_noise_map = plot_setting("plot_lens_fit_noise_map")
        self.plot_lens_fit_signal_to_noise_map = plot_setting(
            "plot_lens_fit_signal_to_noise_map"
        )
        self.plot_lens_fit_model_image = plot_setting("plot_lens_fit_model_image")
        self.plot_lens_fit_residual_map = plot_setting("plot_lens_fit_residual_map")
        self.plot_lens_fit_normalized_residual_map = plot_setting(
            "plot_lens_fit_normalized_residual_map"
        )
        self.plot_lens_fit_chi_squared_map = plot_setting(
            "plot_lens_fit_chi_squared_map"
        )
        self.plot_lens_fit_contribution_maps = plot_setting(
            "plot_lens_fit_contribution_maps"
        )
        self.plot_lens_fit_pixelization_residual_map = plot_setting(
            "plot_lens_fit_pixelization_residual_map"
        )
        self.plot_lens_fit_pixelization_normalized_residuals = plot_setting(
            "plot_lens_fit_pixelization_normalized_residual_map"
        )
        self.plot_lens_fit_pixelization_chi_squared_map = plot_setting(
            "plot_lens_fit_pixelization_chi_squared_map"
        )
        self.plot_lens_fit_pixelization_regularization_weights = plot_setting(
            "plot_lens_fit_pixelization_regularization_weight_map"
        )
        self.plot_lens_fit_subtracted_images_of_planes = plot_setting(
            "plot_lens_fit_subtracted_images_of_planes"
        )
        self.plot_lens_fit_model_images_of_planes = plot_setting(
            "plot_lens_fit_model_images_of_planes"
        )
        self.plot_lens_fit_plane_images_of_planes = plot_setting(
            "plot_lens_fit_plane_images_of_planes"
        )
        self.plot_hyper_model_image = plot_setting("plot_hyper_model_image")
        self.plot_hyper_galaxy_images = plot_setting("plot_hyper_galaxy_images")
        self.plot_binned_hyper_galaxy_images = plot_setting(
            "plot_binned_hyper_galaxy_images"
        )
        self.extract_array_from_mask = figure_setting("extract_images_from_mask")
        self.should_plot_positions = figure_setting("plot_positions_on_images")

        self.plot_ray_tracing_as_subplot = plot_setting("plot_ray_tracing_as_subplot")
        self.plot_ray_tracing_profile_image = plot_setting(
            "plot_ray_tracing_profile_image"
        )
        self.plot_ray_tracing_source_plane = plot_setting(
            "plot_ray_tracing_source_plane_image"
        )
        self.plot_ray_tracing_convergence = plot_setting("plot_ray_tracing_convergence")
        self.plot_ray_tracing_potential = plot_setting("plot_ray_tracing_potential")
        self.plot_ray_tracing_deflections = plot_setting("plot_ray_tracing_deflections")
        self.plot_ray_tracing_magnification = plot_setting(
            "plot_ray_tracing_magnification"
        )

    def plot_ray_tracing(self, tracer, during_analysis):
        positions = (
            self.lens_imaging_data.positions if self.should_plot_positions else None
        )
        mask = self.lens_imaging_data.mask if self.should_plot_mask else None
        phase_plotters.plot_ray_tracing_for_phase(
            tracer=tracer,
            grid=self.lens_imaging_data.grid,
            during_analysis=during_analysis,
            mask=mask,
            extract_array_from_mask=self.extract_array_from_mask,
            zoom_around_mask=self.zoom_around_mask,
            positions=positions,
            units=self.plot_units,
            should_plot_as_subplot=self.plot_ray_tracing_as_subplot,
            should_plot_all_at_end_png=self.plot_ray_tracing_all_at_end_png,
            should_plot_all_at_end_fits=self.plot_ray_tracing_all_at_end_fits,
            should_plot_image=self.plot_ray_tracing_profile_image,
            should_plot_source_plane=self.plot_ray_tracing_source_plane,
            should_plot_convergence=self.plot_ray_tracing_convergence,
            should_plot_potential=self.plot_ray_tracing_potential,
            should_plot_deflections=self.plot_ray_tracing_deflections,
            visualize_path=self.image_path,
            subplot_path=self.subplot_path,
        )

    def plot_lens_imaging(self, fit, during_analysis):
        positions = (
            self.lens_imaging_data.positions if self.should_plot_positions else None
        )
        phase_plotters.plot_lens_imaging_fit_for_phase(
            fit=fit,
            during_analysis=during_analysis,
            should_plot_mask=self.should_plot_mask,
            extract_array_from_mask=self.extract_array_from_mask,
            zoom_around_mask=self.zoom_around_mask,
            positions=positions,
            should_plot_image_plane_pix=self.should_plot_image_plane_pix,
            should_plot_all_at_end_png=self.plot_lens_fit_all_at_end_png,
            should_plot_all_at_end_fits=self.plot_lens_fit_all_at_end_fits,
            should_plot_fit_as_subplot=self.plot_lens_fit_as_subplot,
            should_plot_fit_of_planes_as_subplot=self.plot_lens_fit_of_planes_as_subplot,
            should_plot_inversion_as_subplot=self.plot_lens_fit_inversion_as_subplot,
            should_plot_image=self.plot_lens_fit_image,
            should_plot_noise_map=self.plot_lens_fit_noise_map,
            should_plot_signal_to_noise_map=self.plot_lens_fit_signal_to_noise_map,
            should_plot_model_image=self.plot_lens_fit_model_image,
            should_plot_residual_map=self.plot_lens_fit_residual_map,
            should_plot_normalized_residual_map=self.plot_lens_fit_normalized_residual_map,
            should_plot_chi_squared_map=self.plot_lens_fit_chi_squared_map,
            should_plot_pixelization_residual_map=self.plot_lens_fit_pixelization_residual_map,
            should_plot_pixelization_normalized_residual_map=self.plot_lens_fit_normalized_residual_map,
            should_plot_pixelization_chi_squared_map=self.plot_lens_fit_pixelization_chi_squared_map,
            should_plot_pixelization_regularization_weights=(
                self.plot_lens_fit_pixelization_regularization_weights
            ),
            should_plot_subtracted_images_of_planes=self.plot_lens_fit_subtracted_images_of_planes,
            should_plot_model_images_of_planes=self.plot_lens_fit_model_images_of_planes,
            should_plot_plane_images_of_planes=self.plot_lens_fit_plane_images_of_planes,
            units=self.plot_units,
            visualize_path=self.image_path,
            subplot_path=self.subplot_path,
        )

    def plot_imaging(self):
        mask = self.lens_imaging_data.mask if self.should_plot_mask else None
        positions = (
            self.lens_imaging_data.positions if self.should_plot_positions else None
        )

        phase_plotters.plot_imaging_for_phase(
            imaging_data=self.lens_imaging_data.imaging_data,
            mask=mask,
            positions=positions,
            extract_array_from_mask=self.extract_array_from_mask,
            zoom_around_mask=self.zoom_around_mask,
            units=self.plot_units,
            should_plot_as_subplot=self.plot_data_as_subplot,
            should_plot_image=self.plot_data_image,
            should_plot_noise_map=self.plot_data_noise_map,
            should_plot_psf=self.plot_data_psf,
            should_plot_signal_to_noise_map=self.plot_data_signal_to_noise_map,
            should_plot_absolute_signal_to_noise_map=self.plot_data_absolute_signal_to_noise_map,
            should_plot_potential_chi_squared_map=self.plot_data_potential_chi_squared_map,
            visualize_path=self.image_path,
            subplot_path=self.subplot_path,
        )

    def plot_hyper_images(self, last_results):
        mask = self.lens_imaging_data.mask
        if self.should_plot_mask and mask is not None and last_results is not None:
            phase_plotters.plot_hyper_images_for_phase(
                hyper_model_image_2d=mask.mapping.scaled_array_2d_from_array_1d(
                    array_1d=last_results.hyper_model_image_1d
                ),
                hyper_galaxy_image_2d_path_dict=last_results.hyper_galaxy_image_2d_path_dict,
                binned_hyper_galaxy_image_2d_path_dict=last_results.binned_hyper_galaxy_image_2d_path_dict(
                    binned_grid=self.lens_imaging_data.grid.binned
                ),
                mask=self.lens_imaging_data.mask,
                binned_grid=self.lens_imaging_data.grid.binned,
                extract_array_from_mask=self.extract_array_from_mask,
                zoom_around_mask=self.zoom_around_mask,
                units=self.plot_units,
                should_plot_hyper_model_image=self.plot_hyper_model_image,
                should_plot_hyper_galaxy_images=self.plot_hyper_galaxy_images,
                should_plot_binned_hyper_galaxy_images=self.plot_binned_hyper_galaxy_images,
                visualize_path=self.image_path,
            )


class HyperGalaxyVisualizer(SubPlotVisualizer):
    def __init__(self, image_path):
        super().__init__(image_path)
        self.plot_hyper_galaxy_subplot = plot_setting("plot_hyper_galaxy_subplot")

    def hyper_galaxy_subplot(
        self,
        hyper_galaxy_image,
        contribution_map,
        noise_map,
        hyper_noise_map,
        chi_squared_map,
        hyper_chi_squared_map,
    ):
        hyper_plotters.plot_hyper_galaxy_subplot(
            hyper_galaxy_image=hyper_galaxy_image,
            contribution_map=contribution_map,
            noise_map=noise_map,
            hyper_noise_map=hyper_noise_map,
            chi_squared_map=chi_squared_map,
            hyper_chi_squared_map=hyper_chi_squared_map,
            output_path=self.subplot_path,
            output_format="png",
        )



import autofit as af
import matplotlib

backend = af.conf.instance.visualize.get("figures", "backend", str)
matplotlib.use(backend)
from matplotlib import pyplot as plt

from autolens import exc
from autolens.plotters import plotter_util, array_plotters
from autolens.lens.plotters import lens_plotter_util


def plot_fit_subplot(
    fit,
    should_plot_mask=True,
    extract_array_from_mask=False,
    zoom_around_mask=False,
    positions=None,
    units="arcsec",
    kpc_per_arcsec=None,
    figsize=None,
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    titlesize=10,
    xlabelsize=10,
    ylabelsize=10,
    xyticksize=10,
    mask_pointsize=10,
    position_pointsize=10.0,
    grid_pointsize=1,
    output_path=None,
    output_filename="galaxy_fit",
    output_format="show",
):

    rows, columns, figsize_tool = plotter_util.get_subplot_rows_columns_figsize(
        number_subplots=4
    )

    mask = lens_plotter_util.get_mask(fit=fit, should_plot_mask=should_plot_mask)

    if figsize is None:
        figsize = figsize_tool

    plt.figure(figsize=figsize)
    plt.subplot(rows, columns, 1)

    plot_galaxy_data_array(
        galaxy_data=fit.galaxy_data,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        positions=positions,
        as_subplot=True,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        grid_pointsize=grid_pointsize,
        position_pointsize=position_pointsize,
        mask_pointsize=mask_pointsize,
        output_path=output_path,
        output_filename=output_filename,
        output_format=output_format,
    )

    plt.subplot(rows, columns, 2)

    lens_plotter_util.plot_model_data(
        fit=fit,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        positions=positions,
        as_subplot=True,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title="Model Galaxy",
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        position_pointsize=position_pointsize,
        mask_pointsize=mask_pointsize,
        output_path=output_path,
        output_filename="",
        output_format=output_format,
    )

    plt.subplot(rows, columns, 3)

    lens_plotter_util.plot_residual_map(
        fit=fit,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        as_subplot=True,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        position_pointsize=position_pointsize,
        mask_pointsize=mask_pointsize,
        output_path=output_path,
        output_filename="",
        output_format=output_format,
    )

    plt.subplot(rows, columns, 4)

    lens_plotter_util.plot_chi_squared_map(
        fit=fit,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        as_subplot=True,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        position_pointsize=position_pointsize,
        mask_pointsize=mask_pointsize,
        output_path=output_path,
        output_filename="",
        output_format=output_format,
    )

    plotter_util.output_subplot_array(
        output_path=output_path,
        output_filename=output_filename,
        output_format=output_format,
    )

    plt.close()


def plot_fit_individuals(
    fit,
    should_plot_mask=True,
    extract_array_from_mask=False,
    zoom_around_mask=False,
    positions=None,
    should_plot_image=False,
    should_plot_noise_map=False,
    should_plot_model_image=False,
    should_plot_residual_map=False,
    should_plot_chi_squared_map=False,
    units="kpc",
    output_path=None,
    output_format="show",
):

    mask = lens_plotter_util.get_mask(fit=fit, should_plot_mask=should_plot_mask)

    kpc_per_arcsec = None

    if should_plot_image:

        plot_galaxy_data_array(
            galaxy_data=fit.galaxy_data,
            mask=mask,
            extract_array_from_mask=extract_array_from_mask,
            zoom_around_mask=zoom_around_mask,
            positions=positions,
            units=units,
            kpc_per_arcsec=kpc_per_arcsec,
            output_path=output_path,
            output_format=output_format,
        )

    if should_plot_noise_map:

        lens_plotter_util.plot_noise_map(
            fit=fit,
            mask=mask,
            extract_array_from_mask=extract_array_from_mask,
            zoom_around_mask=zoom_around_mask,
            positions=positions,
            units=units,
            kpc_per_arcsec=kpc_per_arcsec,
            output_path=output_path,
            output_format=output_format,
        )

    if should_plot_model_image:

        lens_plotter_util.plot_model_data(
            fit=fit,
            mask=mask,
            extract_array_from_mask=extract_array_from_mask,
            zoom_around_mask=zoom_around_mask,
            positions=positions,
            units=units,
            kpc_per_arcsec=kpc_per_arcsec,
            output_path=output_path,
            output_format=output_format,
        )

    if should_plot_residual_map:

        lens_plotter_util.plot_residual_map(
            fit=fit,
            mask=mask,
            extract_array_from_mask=extract_array_from_mask,
            zoom_around_mask=zoom_around_mask,
            units=units,
            kpc_per_arcsec=kpc_per_arcsec,
            output_path=output_path,
            output_format=output_format,
        )

    if should_plot_chi_squared_map:

        lens_plotter_util.plot_chi_squared_map(
            fit=fit,
            mask=mask,
            extract_array_from_mask=extract_array_from_mask,
            zoom_around_mask=zoom_around_mask,
            units=units,
            kpc_per_arcsec=kpc_per_arcsec,
            output_path=output_path,
            output_format=output_format,
        )


def plot_galaxy_data_array(
    galaxy_data,
    mask=None,
    extract_array_from_mask=False,
    zoom_around_mask=False,
    positions=None,
    as_subplot=False,
    units="arcsec",
    kpc_per_arcsec=None,
    figsize=None,
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    titlesize=10,
    xlabelsize=10,
    ylabelsize=10,
    xyticksize=10,
    mask_pointsize=10,
    position_pointsize=10.0,
    grid_pointsize=1,
    output_path=None,
    output_filename="galaxy_data",
    output_format="show",
):

    if galaxy_data.use_image:
        title = "Galaxy Data Image"
    elif galaxy_data.use_convergence:
        title = "Galaxy Data Convergence"
    elif galaxy_data.use_potential:
        title = "Galaxy Data Potential"
    elif galaxy_data.use_deflections_y:
        title = "Galaxy Data Deflections (y)"
    elif galaxy_data.use_deflections_x:
        title = "Galaxy Data Deflections (x)"
    else:
        raise exc.PlottingException(
            "The galaxy data_type array does not have a True use_profile_type"
        )

    array_plotters.plot_array(
        array=galaxy_data.image(return_in_2d=True),
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        positions=positions,
        as_subplot=as_subplot,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        position_pointsize=position_pointsize,
        grid_pointsize=grid_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )



from test.integration.tests.lens__source import lens_light_mass__source
from test.integration.tests.lens__source import lens_light_mass__source__hyper
from test.integration.tests.lens__source import lens_light_mass__source__hyper_bg
from test.integration.tests.lens__source import lens_mass__source_x2
from test.integration.tests.lens__source import lens_mass__source_x2__hyper
from test.integration.tests.runner import run_a_mock


class TestCase:
    def _test_lens_light_mass__source(self):
        run_a_mock(lens_light_mass__source)

    def _test_lens_light_mass__source__hyper(self):
        run_a_mock(lens_light_mass__source__hyper)

    def _test_lens_light_mass__source__hyper_bg(self):
        run_a_mock(lens_light_mass__source__hyper_bg)

    def _test_lens_mass__source_x2(self):
        run_a_mock(lens_mass__source_x2)

    def _test_lens_mass__source_x2__hyper(self):
        run_a_mock(lens_mass__source_x2__hyper)



import autolens as al


class MockConvolver(al.Convolver):
    def __init__(self, mask, psf, blurring_mask=None):
        super(MockConvolver, self).__init__(
            mask=mask, blurring_mask=blurring_mask, psf=psf
        )

    def convolver_with_blurring_mask_added(self, blurring_mask):
        return MockConvolver(mask=self.mask, psf=self.psf, blurring_mask=blurring_mask)



import autolens as al

import numpy as np
import pytest


class TestMapping:
    def test__mask_1d_index_tomask_index__compare_to_array_util(self):

        mask = np.array([[True, True, True], [True, False, False], [True, True, False]])

        mask = al.Mask(array=mask, pixel_scale=7.0, sub_size=1)

        mapping = al.Mapping(mask=mask)

        mask_1d_index_tomask_index = al.mask_mapping_util.sub_mask_1d_index_to_submask_index_from_mask_and_sub_size(
            mask=mask, sub_size=1
        )

        assert mapping.mask_1d_index_tomask_index == pytest.approx(
            mask_1d_index_tomask_index, 1e-4
        )

    def test__array_1d_from_array_2d__compare_to_array_util(self):
        array_2d = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9], [10, 11, 12]])

        mask = np.array(
            [
                [True, False, True],
                [False, False, False],
                [True, False, True],
                [True, True, True],
            ]
        )

        array_1d_util = al.array_mapping_util.sub_array_1d_from_sub_array_2d_mask_and_sub_size(
            mask=mask, sub_array_2d=array_2d, sub_size=1
        )

        mask = al.Mask(array=mask, pixel_scale=3.0, sub_size=1)

        mapping = al.Mapping(mask=mask)

        array_1d = mapping.array_1d_from_array_2d(array_2d=array_2d)

        assert (array_1d == array_1d_util).all()

    def test__scaled_array_2d_from_array_1d__compare_to_util(self):
        mask = np.array(
            [
                [True, True, False, False],
                [True, False, True, True],
                [True, True, False, False],
            ]
        )

        array_1d = np.array([1.0, 6.0, 4.0, 5.0, 2.0])

        array_2d_util = al.array_mapping_util.sub_array_2d_from_sub_array_1d_mask_and_sub_size(
            sub_array_1d=array_1d, mask=mask, sub_size=1
        )

        mask = al.Mask(array=mask, pixel_scale=3.0, sub_size=1)

        mapping = al.Mapping(mask=mask)

        array_2d = mapping.array_2d_from_array_1d(array_1d=array_1d)

        assert (array_2d == array_2d_util).all()

        scaled_array_2d = mapping.scaled_array_2d_from_array_1d(array_1d=array_1d)

        assert (scaled_array_2d == array_2d_util).all()
        assert (scaled_array_2d.xticks == np.array([-6.0, -2.0, 2.0, 6.0])).all()
        assert (scaled_array_2d.yticks == np.array([-4.5, -1.5, 1.5, 4.5])).all()
        assert scaled_array_2d.shape_arcsec == (9.0, 12.0)
        assert scaled_array_2d.pixel_scale == 3.0
        assert scaled_array_2d.origin == (0.0, 0.0)

    def test__grid_2d_from_grid_1d__compare_to_util(self):
        mask = np.array(
            [
                [True, True, False, False],
                [True, False, True, True],
                [True, True, False, False],
            ]
        )

        grid_1d = np.array([[1.0, 1.0], [6.0, 6.0], [4.0, 4.0], [5.0, 5.0], [2.0, 2.0]])

        grid_2d_util = al.grid_mapping_util.sub_grid_2d_from_sub_grid_1d_mask_and_sub_size(
            sub_grid_1d=grid_1d, mask=mask, sub_size=1
        )

        mask = al.Mask(array=mask, pixel_scale=2.0, sub_size=1)

        mapping = al.Mapping(mask=mask)
        grid_2d = mapping.grid_2d_from_grid_1d(grid_1d=grid_1d)

        assert (grid_2d_util == grid_2d).all()

    def test__grid_1d_from_grid_2d__compare_to_util(self):
        grid_2d = np.array(
            [
                [[1, 1], [2, 2], [3, 3], [4, 4]],
                [[5, 5], [6, 6], [7, 7], [8, 8]],
                [[9, 9], [10, 10], [11, 11], [12, 12]],
            ]
        )

        mask = np.array(
            [
                [True, False, True, True],
                [False, False, False, True],
                [True, False, True, False],
            ]
        )

        grid_1d_util = al.grid_mapping_util.sub_grid_1d_from_sub_grid_2d_mask_and_sub_size(
            sub_grid_2d=grid_2d, mask=mask, sub_size=1
        )

        mask = al.Mask(array=mask, pixel_scale=2.0, sub_size=1)

        mapping = al.Mapping(mask=mask)
        grid_1d = mapping.grid_1d_from_grid_2d(grid_2d=grid_2d)

        assert (grid_1d_util == grid_1d).all()

    def test__sub_mask_1d_index_to_submask_index__compare_to_array_util(self):
        mask = np.array([[True, True, True], [True, False, False], [True, True, False]])

        mask = al.Mask(array=mask, pixel_scale=7.0, sub_size=2)

        mapping = al.Mapping(mask=mask)

        sub_mask_1d_index_to_submask_index = al.mask_mapping_util.sub_mask_1d_index_to_submask_index_from_mask_and_sub_size(
            mask=mask, sub_size=2
        )

        assert mapping.sub_mask_1d_index_to_submask_index == pytest.approx(
            sub_mask_1d_index_to_submask_index, 1e-4
        )

    def test__sub_array_2d_from_sub_array_1d__use_real_mask_and_grid(self):
        mask = np.array([[False, True], [False, False]])
        mask = al.Mask(array=mask, pixel_scale=3.0, sub_size=2)
        mapping = al.Mapping(mask=mask)

        sub_array_1d = np.array(
            [1.0, 2.0, 3.0, 4.0, 9.0, 10.0, 11.0, 12.0, 13.0, 14.0, 15.0, 16.0]
        )

        sub_array_2d = mapping.sub_array_2d_from_sub_array_1d(sub_array_1d=sub_array_1d)

        assert (
            sub_array_2d
            == np.array(
                [
                    [1.0, 2.0, 0.0, 0.0],
                    [3.0, 4.0, 0.0, 0.0],
                    [9.0, 10.0, 13.0, 14.0],
                    [11.0, 12.0, 15.0, 16.0],
                ]
            )
        ).all()

    def test__sub_array_2d_from_sub_array_1d__use_2x3_mask(self):
        mask = np.array([[False, False, True], [False, True, False]])
        mask = al.Mask(array=mask, pixel_scale=3.0, sub_size=2)
        mapping = al.Mapping(mask=mask)

        sub_array_1d = np.array(
            [
                1.0,
                1.0,
                1.0,
                1.0,
                2.0,
                2.0,
                2.0,
                2.0,
                3.0,
                3.0,
                3.0,
                3.0,
                4.0,
                4.0,
                4.0,
                4.0,
            ]
        )

        sub_array_2d = mapping.sub_array_2d_from_sub_array_1d(sub_array_1d=sub_array_1d)

        assert (
            sub_array_2d
            == np.array(
                [
                    [1.0, 1.0, 2.0, 2.0, 0.0, 0.0],
                    [1.0, 1.0, 2.0, 2.0, 0.0, 0.0],
                    [3.0, 3.0, 0.0, 0.0, 4.0, 4.0],
                    [3.0, 3.0, 0.0, 0.0, 4.0, 4.0],
                ]
            )
        ).all()

    def test__scaled_sub_array_2d_from_sub_array_1d(self):
        mask = np.array([[False, False, True], [False, True, False]])
        mask = al.Mask(array=mask, pixel_scale=3.0, sub_size=2)
        mapping = al.Mapping(mask=mask)

        sub_array_1d = np.array(
            [
                1.0,
                1.0,
                1.0,
                1.0,
                2.0,
                2.0,
                2.0,
                2.0,
                3.0,
                3.0,
                3.0,
                3.0,
                4.0,
                4.0,
                4.0,
                4.0,
            ]
        )

        scaled_sub_array_2d = mapping.scaled_array_2d_with_sub_dimensions_from_sub_array_1d(
            sub_array_1d=sub_array_1d
        )

        assert (
            scaled_sub_array_2d
            == np.array(
                [
                    [1.0, 1.0, 2.0, 2.0, 0.0, 0.0],
                    [1.0, 1.0, 2.0, 2.0, 0.0, 0.0],
                    [3.0, 3.0, 0.0, 0.0, 4.0, 4.0],
                    [3.0, 3.0, 0.0, 0.0, 4.0, 4.0],
                ]
            )
        ).all()

        assert scaled_sub_array_2d.pixel_scales == (1.5, 1.5)
        assert scaled_sub_array_2d.origin == (0.0, 0.0)

    def test__scaled_array_from_sub_array_1d_by_binning_up(self):
        mask = np.array([[False, False, True], [False, True, False]])
        mask = al.Mask(array=mask, pixel_scale=3.0, sub_size=2)
        mapping = al.Mapping(mask=mask)

        sub_array_1d = np.array(
            [
                1.0,
                10.0,
                2.0,
                1.0,
                2.0,
                2.0,
                2.0,
                2.0,
                3.0,
                3.0,
                3.0,
                3.0,
                4.0,
                0.0,
                0.0,
                4.0,
            ]
        )

        scaled_array_2d = mapping.scaled_array_2d_binned_from_sub_array_1d(
            sub_array_1d=sub_array_1d
        )

        assert (scaled_array_2d == np.array([[3.5, 2.0, 0.0], [3.0, 0.0, 2.0]])).all()

        assert scaled_array_2d.pixel_scales == (3.0, 3.0)
        assert scaled_array_2d.origin == (0.0, 0.0)

    def test__grid_1d_binned_from_sub_grid_1d__use_real_mask_and_grid(self):
        mask = np.array([[False, True], [False, False]])
        mask = al.Mask(array=mask, pixel_scale=3.0, sub_size=2)
        mapping = al.Mapping(mask=mask)

        grid_1d = np.array(
            [
                [1.0, 1.0],
                [2.0, 2.0],
                [3.0, 3.0],
                [4.0, 6.0],
                [9.0, 9.0],
                [10.0, 10.0],
                [11.0, 11.0],
                [12.0, 12.0],
                [13.0, 13.0],
                [14.0, 14.0],
                [15.0, 15.0],
                [16.0, 16.0],
            ]
        )

        grid_1d_binned = mapping.grid_1d_binned_from_sub_grid_1d(sub_grid_1d=grid_1d)

        assert (
            grid_1d_binned == np.array([[2.5, 3.0], [10.5, 10.5], [14.5, 14.5]])
        ).all()

    def test__grid_2d_binned_from_sub_grid_1d__use_real_mask_and_grid(self):
        mask = np.array([[False, True], [False, False]])
        mask = al.Mask(array=mask, pixel_scale=3.0, sub_size=2)
        mapping = al.Mapping(mask=mask)

        sub_grid_1d = np.array(
            [
                [1.0, 1.0],
                [2.0, 2.0],
                [3.0, 3.0],
                [4.0, 6.0],
                [9.0, 9.0],
                [10.0, 10.0],
                [11.0, 11.0],
                [12.0, 12.0],
                [13.0, 13.0],
                [14.0, 14.0],
                [15.0, 15.0],
                [16.0, 16.0],
            ]
        )

        grid_2d_binned = mapping.grid_2d_binned_from_sub_grid_1d(
            sub_grid_1d=sub_grid_1d
        )

        assert (
            grid_2d_binned
            == np.array([[[2.5, 3.0], [0.0, 0.0]], [[10.5, 10.5], [14.5, 14.5]]])
        ).all()

    def test__sub_grid_1d_with_sub_dimensions_from_sub_grid_2d__use_real_mask_and_grid(
        self
    ):
        mask = np.array([[False, True], [False, False]])
        mask = al.Mask(array=mask, pixel_scale=2.0, sub_size=2)
        mapping = al.Mapping(mask=mask)

        sub_grid_2d = np.array(
            [
                [[1.0, 1.0], [2.0, 2.0], [-1.0, -1.0], [-1.0, -1.0]],
                [[3.0, 3.0], [4.0, 4.0], [-1.0, -1.0], [-1.0, -1.0]],
                [[5.0, 5.0], [6.0, 6.0], [9.0, 9.0], [10.0, 10.0]],
                [[7.0, 7.0], [8.0, 8.0], [11.0, 11.0], [12.0, 12.0]],
            ]
        )

        sub_grid_1d = mapping.sub_grid_1d_with_sub_dimensions_from_sub_grid_2d(
            sub_grid_2d=sub_grid_2d
        )

        assert (
            sub_grid_1d
            == np.array(
                [
                    [1.0, 1.0],
                    [2.0, 2.0],
                    [3.0, 3.0],
                    [4.0, 4.0],
                    [5.0, 5.0],
                    [6.0, 6.0],
                    [7.0, 7.0],
                    [8.0, 8.0],
                    [9.0, 9.0],
                    [10.0, 10.0],
                    [11.0, 11.0],
                    [12.0, 12.0],
                ]
            )
        ).all()

    def test__sub_grid_2d_with_sub_dimensions_from_sub_grid_1d__use_real_mask_and_grid(
        self
    ):
        mask = np.array([[False, True], [False, False]])
        mask = al.Mask(array=mask, pixel_scale=3.0, sub_size=2)
        mapping = al.Mapping(mask=mask)

        sub_grid_1d = np.array(
            [
                [1.0, 1.0],
                [2.0, 2.0],
                [3.0, 3.0],
                [4.0, 4.0],
                [9.0, 9.0],
                [10.0, 10.0],
                [11.0, 11.0],
                [12.0, 12.0],
                [13.0, 13.0],
                [14.0, 14.0],
                [15.0, 15.0],
                [16.0, 16.0],
            ]
        )

        sub_grid_2d = mapping.sub_grid_2d_with_sub_dimensions_from_sub_grid_1d(
            sub_grid_1d=sub_grid_1d
        )

        assert (
            sub_grid_2d
            == np.array(
                [
                    [[1.0, 1.0], [2.0, 2.0], [0.0, 0.0], [0.0, 0.0]],
                    [[3.0, 3.0], [4.0, 4.0], [0.0, 0.0], [0.0, 0.0]],
                    [[9.0, 9.0], [10.0, 10.0], [13.0, 13.0], [14.0, 14.0]],
                    [[11.0, 11.0], [12.0, 12.0], [15.0, 15.0], [16.0, 16.0]],
                ]
            )
        ).all()

    def test__sub_grid_2d_from_sub_grid_1d__use_2x3_mask(self):
        mask = np.array([[False, False, True], [False, True, False]])
        mask = al.Mask(array=mask, pixel_scale=3.0, sub_size=2)
        mapping = al.Mapping(mask=mask)

        sub_grid_1d = np.array(
            [
                [1.0, 1.0],
                [1.0, 1.0],
                [1.0, 1.0],
                [1.0, 1.0],
                [2.0, 2.0],
                [2.0, 2.0],
                [2.0, 2.0],
                [2.0, 2.0],
                [3.0, 3.0],
                [3.0, 3.0],
                [3.0, 3.0],
                [3.0, 3.0],
                [4.0, 4.0],
                [4.0, 4.0],
                [4.0, 4.0],
                [4.0, 4.0],
            ]
        )

        sub_grid_2d = mapping.sub_grid_2d_with_sub_dimensions_from_sub_grid_1d(
            sub_grid_1d=sub_grid_1d
        )

        assert (
            sub_grid_2d
            == np.array(
                [
                    [
                        [1.0, 1.0],
                        [1.0, 1.0],
                        [2.0, 2.0],
                        [2.0, 2.0],
                        [0.0, 0.0],
                        [0.0, 0.0],
                    ],
                    [
                        [1.0, 1.0],
                        [1.0, 1.0],
                        [2.0, 2.0],
                        [2.0, 2.0],
                        [0.0, 0.0],
                        [0.0, 0.0],
                    ],
                    [
                        [3.0, 3.0],
                        [3.0, 3.0],
                        [0.0, 0.0],
                        [0.0, 0.0],
                        [4.0, 4.0],
                        [4.0, 4.0],
                    ],
                    [
                        [3.0, 3.0],
                        [3.0, 3.0],
                        [0.0, 0.0],
                        [0.0, 0.0],
                        [4.0, 4.0],
                        [4.0, 4.0],
                    ],
                ]
            )
        ).all()

    def test__sub_array_1d_from_sub_array_2d__numerical_values(self):
        sub_array_2d = np.array(
            [
                [1.0, 1.0, 2.0, 2.0, 0.0, 0.0],
                [1.0, 1.0, 2.0, 2.0, 0.0, 0.0],
                [3.0, 3.0, 0.0, 0.0, 4.0, 4.0],
                [3.0, 3.0, 0.0, 0.0, 4.0, 4.0],
            ]
        )

        mask = np.array([[False, False, True], [False, True, False]])
        mask = al.Mask(array=mask, pixel_scale=2.0, sub_size=2)
        mapping = al.Mapping(mask=mask)

        sub_array_1d = mapping.sub_array_1d_with_sub_dimensions_from_sub_array_2d(
            sub_array_2d=sub_array_2d
        )

        assert (
            sub_array_1d
            == np.array(
                [
                    1.0,
                    1.0,
                    1.0,
                    1.0,
                    2.0,
                    2.0,
                    2.0,
                    2.0,
                    3.0,
                    3.0,
                    3.0,
                    3.0,
                    4.0,
                    4.0,
                    4.0,
                    4.0,
                ]
            )
        ).all()

    def test__sub_array_1d_from_sub_array_2d__compare_to_util(self):
        sub_array_2d = np.array(
            [
                [1.0, 1.0, 2.0, 2.0, 3.0, 10.0],
                [1.0, 1.0, 2.0, 2.0, 3.0, 10.0],
                [3.0, 3.0, 8.0, 1.0, 4.0, 4.0],
                [3.0, 3.0, 7.0, 2.0, 4.0, 4.0],
            ]
        )

        mask = np.array([[False, False, False], [True, True, False]])
        mask = al.Mask(array=mask, pixel_scale=2.0, sub_size=2)
        mapping = al.Mapping(mask=mask)

        sub_array_1d = mapping.sub_array_1d_with_sub_dimensions_from_sub_array_2d(
            sub_array_2d=sub_array_2d
        )

        sub_array_1d_util = al.array_mapping_util.sub_array_1d_from_sub_array_2d_mask_and_sub_size(
            sub_array_2d=sub_array_2d, mask=mask, sub_size=2
        )

        assert (sub_array_1d == sub_array_1d_util).all()

    def test__sub_array_to_1d_and_2d_and_back__returns_original_array(self):
        mask = np.array([[False, False, True], [False, True, False]])
        mask = al.Mask(array=mask, pixel_scale=3.0, sub_size=2)
        mapping = al.Mapping(mask=mask)

        sub_array_1d = np.array(
            [
                1.0,
                10.0,
                2.0,
                1.0,
                2.0,
                2.0,
                2.0,
                2.0,
                3.0,
                3.0,
                3.0,
                3.0,
                4.0,
                0.0,
                0.0,
                4.0,
            ]
        )

        sub_array_2d = mapping.sub_array_2d_from_sub_array_1d(sub_array_1d=sub_array_1d)

        sub_array_1d_new = mapping.sub_array_1d_with_sub_dimensions_from_sub_array_2d(
            sub_array_2d=sub_array_2d
        )

        assert (sub_array_1d == sub_array_1d_new).all()

    def test__sub_mask_1d_index_to_mask_1d_index__compare_to_util(self):
        mask = np.array(
            [[True, False, True], [False, False, False], [True, False, False]]
        )

        sub_mask_1d_index_to_mask_1d_index_util = al.mask_mapping_util.sub_mask_1d_index_to_mask_1d_index_from_mask(
            mask=mask, sub_size=2
        )
        mask = al.Mask(array=mask, pixel_scale=3.0, sub_size=2)
        mapping = al.Mapping(mask=mask)

        assert (
            mapping.sub_mask_1d_index_to_mask_1d_index
            == sub_mask_1d_index_to_mask_1d_index_util
        ).all()


class MockObject(object):
    def __init__(self, values):
        self.values = values

    @al.mapping.reshape_returned_sub_array
    def array_from_grid(
        self, grid, return_in_2d=True, return_binned=True, bypass_decorator=False
    ):
        return self.values

    @al.mapping.reshape_returned_grid
    def grid_from_grid(
        self, grid, return_in_2d=True, return_binned=True, bypass_decorator=False
    ):
        return self.values


class TestMappingArrayDecorator(object):
    def test__array_1d_from_function__decorator_changes_array_dimensions_depending_on_inputs(
        self
    ):

        grid = al.Grid.from_shape_pixel_scale_and_sub_size(
            shape=(2, 2), pixel_scale=0.1, sub_size=1
        )

        obj = MockObject(values=np.ones(4))

        array_via_decorator = obj.array_from_grid(
            grid=grid, return_in_2d=False, return_binned=False
        )

        assert (array_via_decorator == np.ones(4)).all()

        array_via_decorator = obj.array_from_grid(
            grid=grid, values=np.ones(4), return_in_2d=True, return_binned=False
        )

        assert (array_via_decorator == np.ones((2, 2))).all()

        obj = MockObject(values=np.ones(16))

        grid = al.Grid.from_shape_pixel_scale_and_sub_size(
            shape=(2, 2), pixel_scale=0.1, sub_size=2
        )

        array_via_decorator = obj.array_from_grid(
            grid=grid, values=np.ones(16), return_in_2d=False, return_binned=False
        )

        assert (array_via_decorator == np.ones(16)).all()

        array_via_decorator = obj.array_from_grid(
            grid=grid, values=np.ones(16), return_in_2d=True, return_binned=False
        )

        assert (array_via_decorator == np.ones((4, 4))).all()

    def test__array_1d_from_function__array_is_binned_if_used__maintains_correct_dimensions(
        self
    ):
        sub_array = np.array(
            [
                1.0,
                1.0,
                1.0,
                1.0,
                2.0,
                2.0,
                2.0,
                2.0,
                3.0,
                3.0,
                3.0,
                3.0,
                4.0,
                4.0,
                4.0,
                4.0,
            ]
        )

        obj = MockObject(values=sub_array)

        grid = al.Grid.from_shape_pixel_scale_and_sub_size(
            shape=(2, 2), pixel_scale=0.1, sub_size=2
        )

        array_via_decorator = obj.array_from_grid(
            grid=grid, return_in_2d=False, return_binned=False
        )

        assert (array_via_decorator == sub_array).all()

        array_via_decorator = obj.array_from_grid(
            grid=grid, return_in_2d=True, return_binned=False
        )

        assert (
            array_via_decorator
            == np.array(
                [
                    [1.0, 1.0, 2.0, 2.0],
                    [1.0, 1.0, 2.0, 2.0],
                    [3.0, 3.0, 4.0, 4.0],
                    [3.0, 3.0, 4.0, 4.0],
                ]
            )
        ).all()

        array_via_decorator = obj.array_from_grid(
            grid=grid, return_in_2d=False, return_binned=True
        )

        assert (array_via_decorator == np.array([1.0, 2.0, 3.0, 4.0])).all()

        array_via_decorator = obj.array_from_grid(
            grid=grid, return_in_2d=True, return_binned=True, bypass_decorator=False
        )

        assert (array_via_decorator == np.array([[1.0, 2.0], [3.0, 4.0]])).all()

    def test__returned_array_from_function_is_2d__grid_in__decorator_convert_dimensions_to_1d_first(
        self
    ):
        obj = MockObject(values=np.ones((4, 4)))

        grid = al.Grid.from_shape_pixel_scale_and_sub_size(
            shape=(4, 4), pixel_scale=0.1, sub_size=1
        )

        array_via_decorator = obj.array_from_grid(
            grid=grid, return_in_2d=False, return_binned=False
        )

        assert (array_via_decorator == np.array(np.ones(16))).all()

        array_via_decorator = obj.array_from_grid(
            grid=grid, return_in_2d=True, return_binned=False
        )

        assert (array_via_decorator == np.array(np.ones((4, 4)))).all()

    def test__returned_array_from_function_is_2d__sub_grid_in__decorator_converts_dimensions_to_1_first(
        self
    ):
        obj = MockObject(values=np.ones((8, 8)))

        grid = al.Grid.from_shape_pixel_scale_and_sub_size(
            shape=(4, 4), pixel_scale=0.1, sub_size=2
        )

        array_via_decorator = obj.array_from_grid(
            grid=grid, return_in_2d=False, return_binned=False
        )

        assert (array_via_decorator == np.ones(64)).all()

        array_via_decorator = obj.array_from_grid(
            grid=grid, return_in_2d=True, return_binned=False
        )

        assert (array_via_decorator == np.ones((8, 8))).all()

        array_via_decorator = obj.array_from_grid(
            grid=grid, return_in_2d=False, return_binned=True
        )

        assert (array_via_decorator == np.ones(16)).all()

        array_via_decorator = obj.array_from_grid(
            grid=grid, return_in_2d=True, return_binned=True, bypass_decorator=False
        )

        assert (array_via_decorator == np.ones((4, 4))).all()


class TestMappingGridDecorator(object):
    def test__grid_1d_from_function__decorator_changes_grid_dimensions_to_2d(self):
        obj = MockObject(values=np.ones(shape=(4, 2)))

        grid = al.Grid.from_shape_pixel_scale_and_sub_size(
            shape=(2, 2), pixel_scale=0.1, sub_size=1
        )

        grid_via_decorator = obj.grid_from_grid(
            grid=grid, return_in_2d=False, return_binned=False
        )

        assert (grid_via_decorator == np.ones((4, 2))).all()

        grid_via_decorator = obj.grid_from_grid(
            grid=grid, return_in_2d=True, return_binned=False
        )

        assert (grid_via_decorator == np.ones((2, 2, 2))).all()

        obj = MockObject(values=np.ones(shape=(16, 2)))

        grid = al.Grid.from_shape_pixel_scale_and_sub_size(
            shape=(2, 2), pixel_scale=0.1, sub_size=2
        )

        grid_via_decorator = obj.grid_from_grid(
            grid=grid, return_in_2d=False, return_binned=False
        )

        assert (grid_via_decorator == np.ones(shape=(16, 2))).all()

        grid_via_decorator = obj.grid_from_grid(
            grid=grid, return_in_2d=True, return_binned=False
        )

        assert (grid_via_decorator == np.ones(shape=(4, 4, 2))).all()

    def test__grid_1d_from_function__binned_up_function_is_used_if_true__maintains_correct_dimensions(
        self
    ):
        sub_grid_values = np.array(
            [
                [1.0, 1.0],
                [1.0, 1.0],
                [1.0, 1.0],
                [1.0, 1.0],
                [2.0, 2.0],
                [2.0, 2.0],
                [2.0, 2.0],
                [2.0, 2.0],
                [3.0, 3.0],
                [3.0, 3.0],
                [3.0, 3.0],
                [3.0, 3.0],
                [4.0, 4.0],
                [4.0, 4.0],
                [4.0, 4.0],
                [4.0, 4.0],
            ]
        )

        obj = MockObject(values=sub_grid_values)

        grid = al.Grid.from_shape_pixel_scale_and_sub_size(
            shape=(2, 2), pixel_scale=0.1, sub_size=2
        )

        grid_via_decorator = obj.grid_from_grid(
            grid=grid, return_in_2d=False, return_binned=False
        )

        assert (grid_via_decorator == sub_grid_values).all()

        grid_via_decorator = obj.grid_from_grid(
            grid=grid, return_in_2d=True, return_binned=False
        )

        assert (
            grid_via_decorator
            == np.array(
                [
                    [[1.0, 1.0], [1.0, 1.0], [2.0, 2.0], [2.0, 2.0]],
                    [[1.0, 1.0], [1.0, 1.0], [2.0, 2.0], [2.0, 2.0]],
                    [[3.0, 3.0], [3.0, 3.0], [4.0, 4.0], [4.0, 4.0]],
                    [[3.0, 3.0], [3.0, 3.0], [4.0, 4.0], [4.0, 4.0]],
                ]
            )
        ).all()

        grid_via_decorator = obj.grid_from_grid(
            grid=grid, return_in_2d=False, return_binned=True
        )

        assert (
            grid_via_decorator
            == np.array([[1.0, 1.0], [2.0, 2.0], [3.0, 3.0], [4.0, 4.0]])
        ).all()

        grid_via_decorator = obj.grid_from_grid(
            grid=grid, return_in_2d=True, return_binned=True, bypass_decorator=False
        )

        assert (
            grid_via_decorator
            == np.array([[[1.0, 1.0], [2.0, 2.0]], [[3.0, 3.0], [4.0, 4.0]]])
        ).all()

    def test__grid_of_function_output_in_2d__decorator_converts_to_1d_first(self):
        obj = MockObject(values=np.ones(shape=(4, 4, 2)))

        grid = al.Grid.from_shape_pixel_scale_and_sub_size(
            shape=(4, 4), pixel_scale=0.1, sub_size=1
        )

        grid_via_decorator = obj.grid_from_grid(
            grid=grid, return_in_2d=False, return_binned=False
        )

        assert (grid_via_decorator == np.array(np.ones(shape=(16, 2)))).all()

        grid_via_decorator = obj.grid_from_grid(
            grid=grid, return_in_2d=True, return_binned=False
        )

        assert (grid_via_decorator == np.array(np.ones(shape=(4, 4, 2)))).all()

    def test__sub_grid_of_function_returned_in_2d__decorator_changes_grid_dimensions_to_2d(
        self
    ):
        obj = MockObject(values=np.ones(shape=(8, 8, 2)))

        grid = al.Grid.from_shape_pixel_scale_and_sub_size(
            shape=(4, 4), pixel_scale=0.1, sub_size=2
        )

        grid_via_decorator = obj.grid_from_grid(
            grid=grid, return_in_2d=False, return_binned=False
        )

        assert (grid_via_decorator == np.ones((64, 2))).all()

        grid_via_decorator = obj.grid_from_grid(
            grid=grid, return_in_2d=True, return_binned=False
        )

        assert (grid_via_decorator == np.ones((8, 8, 2))).all()

        grid_via_decorator = obj.grid_from_grid(
            grid=grid, return_in_2d=False, return_binned=True
        )

        assert (grid_via_decorator == np.ones((16, 2))).all()

        grid_via_decorator = obj.grid_from_grid(
            grid=grid, return_in_2d=True, return_binned=True, bypass_decorator=False
        )

        assert (grid_via_decorator == np.ones((4, 4, 2))).all()



import logging

import numpy as np

from autolens import decorator_util

logger = logging.getLogger(__name__)
logger.level = logging.DEBUG


@decorator_util.jit()
def sparse_to_unmasked_sparse_from_mask_and_pixel_centres(
    total_sparse_pixels, mask, unmasked_sparse_grid_pixel_centres
):
    """Determine the mapping_util between every masked pixelization-grid pixel and pixelization-grid pixel. This is \
    performed by checking whether each pixelization-grid pixel is within the masks, and mapping_util the indexes.

    Parameters
    -----------
    total_sparse_pixels : int
        The total number of pixels in the pixelization grid which fall within the masks.
    mask : imaging.masks.Mask
        The masks within which pixelization pixels must be inside
    unmasked_sparse_grid_pixel_centres : ndarray
        The centres of the unmasked pixelization grid pixels.
    """

    pix_to_full_pix = np.zeros(total_sparse_pixels)

    pixel_index = 0

    for full_pixel_index in range(unmasked_sparse_grid_pixel_centres.shape[0]):

        y = unmasked_sparse_grid_pixel_centres[full_pixel_index, 0]
        x = unmasked_sparse_grid_pixel_centres[full_pixel_index, 1]

        if not mask[y, x]:
            pix_to_full_pix[pixel_index] = full_pixel_index
            pixel_index += 1

    return pix_to_full_pix


@decorator_util.jit()
def unmasked_sparse_to_sparse_from_mask_and_pixel_centres(
    mask, unmasked_sparse_grid_pixel_centres, total_sparse_pixels
):
    """Determine the mapping_util between every pixelization-grid pixel and masked pixelization-grid pixel. This is \
    performed by checking whether each pixelization-grid pixel is within the masks, and mapping_util the indexes.

    Pixelization pixels are paired with the next masked pixel index. This may mean that a pixel is not paired with a \
    pixel near it, if the next pixel is on the next row of the grid. This is not a problem, as it is only \
    unmasked pixels that are referenced when computing image_to_pix, which is what this array is used for.

    Parameters
    -----------
    total_sparse_pixels : int
        The total number of pixels in the pixelization grid which fall within the masks.
    mask : imaging.masks.Mask
        The masks within which pixelization pixels must be inside
    unmasked_sparse_grid_pixel_centres : ndarray
        The centres of the unmasked pixelization grid pixels.
    """

    total_unmasked_sparse_pixels = unmasked_sparse_grid_pixel_centres.shape[0]

    unmasked_sparse_to_sparse = np.zeros(total_unmasked_sparse_pixels)
    pixel_index = 0

    for unmasked_sparse_pixel_index in range(total_unmasked_sparse_pixels):

        y = unmasked_sparse_grid_pixel_centres[unmasked_sparse_pixel_index, 0]
        x = unmasked_sparse_grid_pixel_centres[unmasked_sparse_pixel_index, 1]

        unmasked_sparse_to_sparse[unmasked_sparse_pixel_index] = pixel_index

        if not mask[y, x]:
            if pixel_index < total_sparse_pixels - 1:
                pixel_index += 1

    return unmasked_sparse_to_sparse


@decorator_util.jit()
def mask_1d_index_to_sparse_1d_index_from_sparse_mappings(
    regular_to_unmasked_sparse, unmasked_sparse_to_sparse
):
    """Using the mapping_util between the grid and unmasked pixelization grid, compute the mapping_util between each \
    pixel and the masked pixelization grid.

    Parameters
    -----------
    regular_to_unmasked_sparse : ndarray
        The index mapping_util between every pixel and masked pixelization pixel.
    unmasked_sparse_to_sparse : ndarray
        The index mapping_util between every masked pixelization pixel and unmasked pixelization pixel.
    """
    total_regular_pixels = regular_to_unmasked_sparse.shape[0]

    mask_1d_index_to_sparse_1d_index = np.zeros(total_regular_pixels)

    for regular_index in range(total_regular_pixels):
        mask_1d_index_to_sparse_1d_index[regular_index] = unmasked_sparse_to_sparse[
            regular_to_unmasked_sparse[regular_index]
        ]

    return mask_1d_index_to_sparse_1d_index


@decorator_util.jit()
def sparse_grid_from_unmasked_sparse_grid(
    unmasked_sparse_grid, sparse_to_unmasked_sparse
):
    """Use the central arc-second coordinate of every unmasked pixelization grid's pixels and mapping_util between each \
    pixelization pixel and unmasked pixelization pixel to compute the central arc-second coordinate of every masked \
    pixelization grid pixel.

    Parameters
    -----------
    unmasked_sparse_grid : ndarray
        The (y,x) arc-second centre of every unmasked pixelization grid pixel.
    sparse_to_unmasked_sparse : ndarray
        The index mapping_util between every pixelization pixel and masked pixelization pixel.
    """
    total_pix_pixels = sparse_to_unmasked_sparse.shape[0]

    pix_grid = np.zeros((total_pix_pixels, 2))

    for pixel_index in range(total_pix_pixels):
        pix_grid[pixel_index, 0] = unmasked_sparse_grid[
            sparse_to_unmasked_sparse[pixel_index], 0
        ]
        pix_grid[pixel_index, 1] = unmasked_sparse_grid[
            sparse_to_unmasked_sparse[pixel_index], 1
        ]

    return pix_grid


@decorator_util.jit()
def mask_1d_index_to_sparse_1d_index_from_binned_grid(
    sparse_labels,
    binned_mask_1d_index_to_mask_1d_indexes,
    binned_mask_1d_index_to_mask_1d_sizes,
    total_unbinned_pixels,
):
    mask_1d_index_to_sparse_1d_index = np.zeros(total_unbinned_pixels)

    for cluster_index in range(binned_mask_1d_index_to_mask_1d_indexes.shape[0]):
        for cluster_count in range(
            binned_mask_1d_index_to_mask_1d_sizes[cluster_index]
        ):
            regular_index = binned_mask_1d_index_to_mask_1d_indexes[
                cluster_index, cluster_count
            ]
            mask_1d_index_to_sparse_1d_index[regular_index] = sparse_labels[
                cluster_index
            ]

    return mask_1d_index_to_sparse_1d_index



import numpy as np
from astropy import cosmology as cosmo
from itertools import count
from skimage import measure

import autofit as af
from autolens import exc, dimensions as dim
from autolens import text_util
from autolens.array.mapping import (
    reshape_returned_sub_array,
    reshape_returned_array,
    reshape_returned_grid,
)
from autolens.model.inversion import pixelizations as pix
from autolens.model.profiles import light_profiles as lp
from autolens.model.profiles import mass_profiles as mp


def is_light_profile(obj):
    return isinstance(obj, lp.LightProfile)


def is_mass_profile(obj):
    return isinstance(obj, mp.MassProfile)


class Galaxy(af.ModelObject):
    """
    @DynamicAttrs
    """

    def __init__(
        self,
        redshift,
        pixelization=None,
        regularization=None,
        hyper_galaxy=None,
        **kwargs
    ):
        """Class representing a galaxy, which is composed of attributes used for fitting hyper_galaxies (e.g. light profiles, \
        mass profiles, pixelizations, etc.).
        
        All *has_* methods retun *True* if galaxy has that attribute, *False* if not.

        Parameters
        ----------
        redshift: float
            The redshift of the galaxy.
        light_profiles: [lp.LightProfile]
            A list of the galaxy's light profiles.
        mass_profiles: [mp.MassProfile]
            A list of the galaxy's mass profiles.
        hyper_galaxy : HyperGalaxy
            The hyper_galaxies-parameters of the hyper_galaxies-galaxy, which is used for performing a hyper_galaxies-analysis on the noise-map.
            
        Attributes
        ----------
        pixelization : inversion.Pixelization
            The pixelization of the galaxy used to reconstruct an observed image using an inversion.
        regularization : inversion.Regularization
            The regularization of the pixel-grid used to reconstruct an observed using an inversion.
        """
        super().__init__()
        self.redshift = redshift

        self.hyper_model_image_1d = None
        self.hyper_galaxy_image_1d = None
        self.binned_hyper_galaxy_image_1d = None

        for name, val in kwargs.items():
            setattr(self, name, val)

        self.pixelization = pixelization
        self.regularization = regularization

        if pixelization is not None and regularization is None:
            raise exc.GalaxyException(
                "If the galaxy has a pixelization, it must also have a regularization."
            )
        if pixelization is None and regularization is not None:
            raise exc.GalaxyException(
                "If the galaxy has a regularization, it must also have a pixelization."
            )

        self.hyper_galaxy = hyper_galaxy

    def __hash__(self):
        return self.id

    @property
    def light_profiles(self):
        return [value for value in self.__dict__.values() if is_light_profile(value)]

    @property
    def mass_profiles(self):
        return [value for value in self.__dict__.values() if is_mass_profile(value)]

    @property
    def has_redshift(self):
        return self.redshift is not None

    @property
    def has_pixelization(self):
        return self.pixelization is not None

    @property
    def has_regularization(self):
        return self.regularization is not None

    @property
    def has_hyper_galaxy(self):
        return self.hyper_galaxy is not None

    @property
    def has_light_profile(self):
        return len(self.light_profiles) > 0

    @property
    def has_mass_profile(self):
        return len(self.mass_profiles) > 0

    @property
    def has_profile(self):
        return len(self.mass_profiles) + len(self.light_profiles) > 0

    @property
    def uses_cluster_inversion(self):
        return type(self.pixelization) is pix.VoronoiBrightnessImage

    def __repr__(self):
        string = "Redshift: {}".format(self.redshift)
        if self.pixelization:
            string += "\nPixelization:\n{}".format(str(self.pixelization))
        if self.regularization:
            string += "\nRegularization:\n{}".format(str(self.regularization))
        if self.hyper_galaxy:
            string += "\nHyper Galaxy:\n{}".format(str(self.hyper_galaxy))
        if self.light_profiles:
            string += "\nLight Profiles:\n{}".format(
                "\n".join(map(str, self.light_profiles))
            )
        if self.mass_profiles:
            string += "\nMass Profiles:\n{}".format(
                "\n".join(map(str, self.mass_profiles))
            )
        return string

    def __eq__(self, other):
        return all(
            (
                isinstance(other, Galaxy),
                self.pixelization == other.pixelization,
                self.redshift == other.redshift,
                self.hyper_galaxy == other.hyper_galaxy,
                self.light_profiles == other.light_profiles,
                self.mass_profiles == other.mass_profiles,
            )
        )

    @reshape_returned_sub_array
    def profile_image_from_grid(
        self, grid, return_in_2d=True, return_binned=True, bypass_decorator=False
    ):
        """Calculate the summed image of all of the galaxy's light profiles using a grid of Cartesian (y,x) \
        coordinates.
        
        If the galaxy has no light profiles, a grid of zeros is returned.
        
        See *profiles.light_profiles* for a description of how light profile image are computed.

        Parameters
        ----------
        grid : ndarray
            The (y, x) coordinates in the original reference frame of the grid.
        return_in_2d : bool
            If *True*, the returned array is mapped to its unmasked 2D shape, if *False* it is the masked 1D shape.
        return_binned : bool
            If *True*, the returned array which is computed on a sub-grid is binned up to the grid dimensions \
            by taking the mean of all sub-gridded values. If *False*, the array is returned on the dimensions of the \
            sub-grid.
        """
        if self.has_light_profile:
            return sum(
                map(
                    lambda p: p.profile_image_from_grid(
                        grid=grid,
                        return_in_2d=False,
                        return_binned=False,
                        bypass_decorator=True,
                    ),
                    self.light_profiles,
                )
            )
        else:
            return np.zeros((grid.shape[0],))

    @reshape_returned_array
    def blurred_profile_image_from_grid_and_psf(
        self, grid, psf, preload_blurring_grid=None, return_in_2d=True
    ):

        profile_image = self.profile_image_from_grid(
            grid=grid, return_in_2d=True, return_binned=True
        )

        if preload_blurring_grid is None:
            preload_blurring_grid = grid.blurring_grid_from_psf_shape(
                psf_shape=psf.shape
            )

        blurring_image = self.profile_image_from_grid(
            grid=preload_blurring_grid, return_in_2d=True, return_binned=True
        )

        return psf.convolve(profile_image + blurring_image)

    @reshape_returned_array
    def blurred_profile_image_from_grid_and_convolver(
        self, grid, convolver, preload_blurring_grid=None, return_in_2d=True
    ):

        if convolver.blurring_mask is None:
            blurring_mask = grid.mask.blurring_mask_from_psf_shape(
                psf_shape=convolver.psf.shape
            )
            convolver = convolver.convolver_with_blurring_mask_added(
                blurring_mask=blurring_mask
            )

        profile_image = self.profile_image_from_grid(
            grid=grid, return_in_2d=False, return_binned=True
        )

        if preload_blurring_grid is None:
            preload_blurring_grid = grid.blurring_grid_from_psf_shape(
                psf_shape=convolver.psf.shape
            )

        blurring_image = self.profile_image_from_grid(
            grid=preload_blurring_grid, return_in_2d=False, return_binned=True
        )

        return convolver.convolve_image(
            image_array=profile_image, blurring_array=blurring_image
        )

    def profile_visibilities_from_grid_and_transformer(self, grid, transformer):

        profile_image_1d = self.profile_image_from_grid(
            grid=grid, return_in_2d=False, return_binned=True
        )

        return transformer.visibilities_from_image_1d(image_1d=profile_image_1d)

    def luminosity_within_circle_in_units(
        self,
        radius: dim.Length,
        unit_luminosity="eps",
        exposure_time=None,
        cosmology=cosmo.Planck15,
        **kwargs
    ):
        """Compute the total luminosity of the galaxy's light profiles within a circle of specified radius.

        See *light_profiles.luminosity_within_circle* for details of how this is performed.

        Parameters
        ----------
        radius : float
            The radius of the circle to compute the dimensionless mass within.
        unit_luminosity : str
            The units the luminosity is returned in (eps | counts).
        exposure_time : float
            The exposure time of the observation, which converts luminosity from electrons per second units to counts.
        """
        if self.has_light_profile:
            return sum(
                map(
                    lambda p: p.luminosity_within_circle_in_units(
                        radius=radius,
                        unit_luminosity=unit_luminosity,
                        redshift_profile=self.redshift,
                        exposure_time=exposure_time,
                        cosmology=cosmology,
                        kwargs=kwargs,
                    ),
                    self.light_profiles,
                )
            )
        return None

    def luminosity_within_ellipse_in_units(
        self,
        major_axis: dim.Length,
        unit_luminosity="eps",
        exposure_time=None,
        cosmology=cosmo.Planck15,
        **kwargs
    ):
        """Compute the total luminosity of the galaxy's light profiles, within an
        ellipse of specified major axis. This is performed via integration of each
        light profile and is centred, oriented and aligned with each light model's
        individual geometry.

        See *light_profiles.luminosity_within_ellipse* for details of how this is \
        performed.

        Parameters
        ----------
        major_axis : float
            The major-axis radius of the ellipse.
        unit_luminosity : str
            The units the luminosity is returned in (eps | counts).
        exposure_time : float
            The exposure time of the observation, which converts luminosity from \
            electrons per second units to counts.
        """
        if self.has_light_profile:
            return sum(
                map(
                    lambda p: p.luminosity_within_ellipse_in_units(
                        major_axis=major_axis,
                        unit_luminosity=unit_luminosity,
                        redshift_profile=self.redshift,
                        exposure_time=exposure_time,
                        cosmology=cosmology,
                        kwargs=kwargs,
                    ),
                    self.light_profiles,
                )
            )
        return None

    @reshape_returned_sub_array
    def convergence_from_grid(
        self, grid, return_in_2d=True, return_binned=True, bypass_decorator=False
    ):
        """Compute the summed convergence of the galaxy's mass profiles using a grid \
        of Cartesian (y,x) coordinates.

        If the galaxy has no mass profiles, a grid of zeros is returned.
        
        See *profiles.mass_profiles* module for details of how this is performed.

        The *reshape_returned_array* decorator reshapes the NumPy array the convergence is outputted on. See \
        *grids.reshape_returned_array* for a description of the output.

        Parameters
        ----------
        grid : ndarray
            The (y, x) coordinates in the original reference frame of the grid.
        return_in_2d : bool
            If *True*, the returned array is mapped to its unmasked 2D shape, if *False* it is the masked 1D shape.
        return_binned : bool
            If *True*, the returned array which is computed on a sub-grid is binned up to the grid dimensions \
            by taking the mean of all sub-gridded values. If *False*, the array is returned on the dimensions of the \
            sub-grid.
        """
        if self.has_mass_profile:
            return sum(
                map(
                    lambda p: p.convergence_from_grid(
                        grid=grid,
                        return_in_2d=False,
                        return_binned=False,
                        bypass_decorator=True,
                    ),
                    self.mass_profiles,
                )
            )
        return np.zeros((grid.shape[0],))

    @reshape_returned_sub_array
    def potential_from_grid(
        self, grid, return_in_2d=True, return_binned=True, bypass_decorator=False
    ):
        """Compute the summed gravitational potential of the galaxy's mass profiles \
        using a grid of Cartesian (y,x) coordinates.

        If the galaxy has no mass profiles, a grid of zeros is returned.

        See *profiles.mass_profiles* module for details of how this is performed.

                The *reshape_returned_array* decorator reshapes the NumPy array the convergence is outputted on. See \
        *grids.reshape_returned_array* for a description of the output.

        Parameters
        ----------
        grid : ndarray
            The (y, x) coordinates in the original reference frame of the grid.
        return_in_2d : bool
            If *True*, the returned array is mapped to its unmasked 2D shape, if *False* it is the masked 1D shape.
        return_binned : bool
            If *True*, the returned array which is computed on a sub-grid is binned up to the grid dimensions \
            by taking the mean of all sub-gridded values. If *False*, the array is returned on the dimensions of the \
            sub-grid.
        """
        if self.has_mass_profile:
            return sum(
                map(
                    lambda p: p.potential_from_grid(
                        grid=grid,
                        return_in_2d=False,
                        return_binned=False,
                        bypass_decorator=True,
                    ),
                    self.mass_profiles,
                )
            )
        return np.zeros((grid.shape[0],))

    @reshape_returned_grid
    def deflections_from_grid(
        self, grid, return_in_2d=True, return_binned=True, bypass_decorator=False
    ):
        """Compute the summed (y,x) deflection angles of the galaxy's mass profiles \
        using a grid of Cartesian (y,x) coordinates.

        If the galaxy has no mass profiles, two grid of zeros are returned.

        See *profiles.mass_profiles* module for details of how this is performed.

        Parameters
        ----------
        grid : ndarray
            The (y, x) coordinates in the original reference frame of the grid.
        """
        if self.has_mass_profile:
            return sum(
                map(
                    lambda p: p.deflections_from_grid(
                        grid=grid,
                        return_in_2d=False,
                        return_binned=False,
                        bypass_decorator=True,
                    ),
                    self.mass_profiles,
                )
            )
        return np.full((grid.shape[0], 2), 0.0)

    @reshape_returned_grid
    def deflections_via_potential_from_grid(
        self, grid, return_in_2d=True, return_binned=True
    ):
        potential_2d = self.potential_from_grid(
            grid=grid, return_in_2d=True, return_binned=False
        )

        deflections_y_2d = np.gradient(potential_2d, grid.in_2d[:, 0, 0], axis=0)
        deflections_x_2d = np.gradient(potential_2d, grid.in_2d[0, :, 1], axis=1)

        return np.stack((deflections_y_2d, deflections_x_2d), axis=-1)

    @reshape_returned_sub_array
    def lensing_jacobian_a11_from_grid(
        self, grid, return_in_2d=True, return_binned=True
    ):

        deflections_2d = self.deflections_from_grid(
            grid=grid, return_in_2d=True, return_binned=False
        )

        return 1.0 - np.gradient(deflections_2d[:, :, 1], grid.in_2d[0, :, 1], axis=1)

    @reshape_returned_sub_array
    def lensing_jacobian_a12_from_grid(
        self, grid, return_in_2d=True, return_binned=True
    ):

        deflections_2d = self.deflections_from_grid(
            grid=grid, return_in_2d=True, return_binned=False
        )

        return -1.0 * np.gradient(deflections_2d[:, :, 1], grid.in_2d[:, 0, 0], axis=0)

    @reshape_returned_sub_array
    def lensing_jacobian_a21_from_grid(
        self, grid, return_in_2d=True, return_binned=True
    ):

        deflections_2d = self.deflections_from_grid(
            grid=grid, return_in_2d=True, return_binned=False
        )

        return -1.0 * np.gradient(deflections_2d[:, :, 0], grid.in_2d[0, :, 1], axis=1)

    @reshape_returned_sub_array
    def lensing_jacobian_a22_from_grid(
        self, grid, return_in_2d=True, return_binned=True
    ):

        deflections_2d = self.deflections_from_grid(
            grid=grid, return_in_2d=True, return_binned=False
        )

        return 1 - np.gradient(deflections_2d[:, :, 0], grid.in_2d[:, 0, 0], axis=0)

    def lensing_jacobian_from_grid(self, grid, return_in_2d=True, return_binned=True):

        a11 = self.lensing_jacobian_a11_from_grid(
            grid=grid, return_in_2d=return_in_2d, return_binned=return_binned
        )

        a12 = self.lensing_jacobian_a12_from_grid(
            grid=grid, return_in_2d=return_in_2d, return_binned=return_binned
        )

        a21 = self.lensing_jacobian_a21_from_grid(
            grid=grid, return_in_2d=return_in_2d, return_binned=return_binned
        )

        a22 = self.lensing_jacobian_a22_from_grid(
            grid=grid, return_in_2d=return_in_2d, return_binned=return_binned
        )

        return np.array([[a11, a12], [a21, a22]])

    @reshape_returned_sub_array
    def convergence_via_jacobian_from_grid(
        self, grid, return_in_2d=True, return_binned=True
    ):

        jacobian = self.lensing_jacobian_from_grid(
            grid=grid, return_in_2d=False, return_binned=False
        )

        convergence = 1 - 0.5 * (jacobian[0, 0] + jacobian[1, 1])

        return convergence

    @reshape_returned_sub_array
    def shear_via_jacobian_from_grid(self, grid, return_in_2d=True, return_binned=True):

        jacobian = self.lensing_jacobian_from_grid(
            grid=grid, return_in_2d=True, return_binned=False
        )

        gamma_1 = 0.5 * (jacobian[1, 1] - jacobian[0, 0])
        gamma_2 = -0.5 * (jacobian[0, 1] + jacobian[1, 0])

        return (gamma_1 ** 2 + gamma_2 ** 2) ** 0.5

    @reshape_returned_sub_array
    def tangential_eigen_value_from_grid(
        self, grid, return_in_2d=True, return_binned=True
    ):

        convergence = self.convergence_via_jacobian_from_grid(
            grid=grid, return_in_2d=False, return_binned=False
        )

        shear = self.shear_via_jacobian_from_grid(
            grid=grid, return_in_2d=False, return_binned=False
        )

        return 1 - convergence - shear

    @reshape_returned_sub_array
    def radial_eigen_value_from_grid(self, grid, return_in_2d=True, return_binned=True):

        convergence = self.convergence_via_jacobian_from_grid(
            grid=grid, return_in_2d=False, return_binned=False
        )

        shear = self.shear_via_jacobian_from_grid(
            grid=grid, return_in_2d=False, return_binned=False
        )

        return 1 - convergence + shear

    @reshape_returned_sub_array
    def magnification_from_grid(self, grid, return_in_2d=True, return_binned=True):

        jacobian = self.lensing_jacobian_from_grid(
            grid=grid, return_in_2d=False, return_binned=False
        )

        det_jacobian = jacobian[0, 0] * jacobian[1, 1] - jacobian[0, 1] * jacobian[1, 0]

        return 1 / det_jacobian

    def tangential_critical_curve_from_grid(self, grid):

        lambda_tangential_2d = self.tangential_eigen_value_from_grid(
            grid=grid, return_in_2d=True, return_binned=False
        )

        tangential_critical_curve_indices = measure.find_contours(
            lambda_tangential_2d, 0
        )

        if len(tangential_critical_curve_indices) == 0:
            return []

        return grid.marching_squares_grid_pixels_to_grid_arcsec(
            grid_pixels=tangential_critical_curve_indices[0],
            shape=lambda_tangential_2d.shape,
        )

    def radial_critical_curve_from_grid(self, grid):

        lambda_radial_2d = self.radial_eigen_value_from_grid(
            grid=grid, return_in_2d=True, return_binned=False
        )

        radial_critical_curve_indices = measure.find_contours(lambda_radial_2d, 0)

        if len(radial_critical_curve_indices) == 0:
            return []

        return grid.marching_squares_grid_pixels_to_grid_arcsec(
            grid_pixels=radial_critical_curve_indices[0], shape=lambda_radial_2d.shape
        )

    def tangential_caustic_from_grid(self, grid):

        tangential_critical_curve = self.tangential_critical_curve_from_grid(grid=grid)

        if len(tangential_critical_curve) == 0:
            return []

        deflections_1d = self.deflections_from_grid(
            grid=tangential_critical_curve, bypass_decorator=True
        )

        return tangential_critical_curve - deflections_1d

    def radial_caustic_from_grid(self, grid):

        radial_critical_curve = self.radial_critical_curve_from_grid(grid=grid)

        if len(radial_critical_curve) == 0:
            return []

        deflections_1d = self.deflections_from_grid(
            grid=radial_critical_curve, bypass_decorator=True
        )

        return radial_critical_curve - deflections_1d

    def critical_curves_from_grid(self, grid):
        return [
            self.tangential_critical_curve_from_grid(grid=grid),
            self.radial_critical_curve_from_grid(grid=grid),
        ]

    def caustics_from_grid(self, grid):
        return [
            self.tangential_caustic_from_grid(grid=grid),
            self.radial_caustic_from_grid(grid=grid),
        ]

    def mass_within_circle_in_units(
        self,
        radius: dim.Length,
        redshift_source=None,
        unit_mass="solMass",
        cosmology=cosmo.Planck15,
        **kwargs
    ):
        """Compute the total angular mass of the galaxy's mass profiles within a \
        circle of specified radius.

        See *profiles.mass_profiles.mass_within_circle* for details of how this is \
        performed.

        Parameters
        ----------
        radius : float
            The radius of the circle to compute the dimensionless mass within.
        unit_mass : str
            The units the mass is returned in (angular | solMass).
        """
        if self.has_mass_profile:
            return sum(
                map(
                    lambda p: p.mass_within_circle_in_units(
                        radius=radius,
                        redshift_profile=self.redshift,
                        redshift_source=redshift_source,
                        unit_mass=unit_mass,
                        cosmology=cosmology,
                        kwargs=kwargs,
                    ),
                    self.mass_profiles,
                )
            )
        return None

    def mass_within_ellipse_in_units(
        self,
        major_axis: dim.Length,
        redshift_source=None,
        unit_mass="solMass",
        cosmology=cosmo.Planck15,
        **kwargs
    ):
        """Compute the total angular mass of the galaxy's mass profiles within an \
        ellipse of specified major_axis.

        See *profiles.mass_profiles.angualr_mass_within_ellipse* for details of how \
        this is performed.

        Parameters
        ----------
        major_axis : float
            The major-axis radius of the ellipse.
        """
        if self.has_mass_profile:
            return sum(
                map(
                    lambda p: p.mass_within_ellipse_in_units(
                        major_axis=major_axis,
                        redshift_profile=self.redshift,
                        redshift_source=redshift_source,
                        unit_mass=unit_mass,
                        cosmology=cosmology,
                        kwargs=kwargs,
                    ),
                    self.mass_profiles,
                )
            )
        return None

    def einstein_radius_in_units(
        self, unit_length="arcsec", cosmology=cosmo.Planck15, **kwargs
    ):
        """The Einstein Radius of this galaxy, which is the sum of Einstein Radii of \
        its mass profiles.

        If the galaxy is composed of multiple elliptical profiles with different \
        axis-ratios, this Einstein Radius may be inaccurate. This is because the \
        differently oriented ellipses of each mass profile """

        if self.has_mass_profile:
            return sum(
                map(
                    lambda p: p.einstein_radius_in_units(
                        unit_length=unit_length,
                        redshift_profile=self.redshift,
                        cosmology=cosmology,
                    ),
                    self.mass_profiles,
                )
            )
        return None

    def einstein_mass_in_units(
        self,
        unit_mass="solMass",
        redshift_source=None,
        cosmology=cosmo.Planck15,
        **kwargs
    ):
        """The Einstein Mass of this galaxy, which is the sum of Einstein Radii of its
        mass profiles.

        If the galaxy is composed of multiple ellipitcal profiles with different \
        axis-ratios, this Einstein Mass may be inaccurate. This is because the \
        differently oriented ellipses of each mass profile """

        if self.has_mass_profile:
            return sum(
                map(
                    lambda p: p.einstein_mass_in_units(
                        unit_mass=unit_mass,
                        redshift_profile=self.redshift,
                        redshift_source=redshift_source,
                        cosmology=cosmology,
                        kwargs=kwargs,
                    ),
                    self.mass_profiles,
                )
            )
        return None

    def summarize_in_units(
        self,
        radii,
        whitespace=80,
        unit_length="arcsec",
        unit_luminosity="eps",
        unit_mass="solMass",
        redshift_source=None,
        cosmology=cosmo.Planck15,
        **kwargs
    ):

        if hasattr(self, "name"):
            summary = ["Galaxy = {}\n".format(self.name)]
            prefix_galaxy = self.name + "_"
        else:
            summary = ["Galaxy\n"]
            prefix_galaxy = ""

        summary += [
            af.text_util.label_and_value_string(
                label=prefix_galaxy + "redshift",
                value=self.redshift,
                whitespace=whitespace,
            )
        ]

        if self.has_light_profile:
            summary += self.summarize_light_profiles_in_units(
                whitespace=whitespace,
                prefix=prefix_galaxy,
                radii=radii,
                unit_length=unit_length,
                unit_luminosity=unit_luminosity,
                redshift_source=redshift_source,
                cosmology=cosmology,
                kwargs=kwargs,
            )

        if self.has_mass_profile:
            summary += self.summarize_mass_profiles_in_units(
                whitespace=whitespace,
                prefix=prefix_galaxy,
                radii=radii,
                unit_length=unit_length,
                unit_mass=unit_mass,
                redshift_source=redshift_source,
                cosmology=cosmology,
                kwargs=kwargs,
            )

        return summary

    def summarize_light_profiles_in_units(
        self,
        radii,
        whitespace=80,
        prefix="",
        unit_length="arcsec",
        unit_luminosity="eps",
        redshift_source=None,
        cosmology=cosmo.Planck15,
        **kwargs
    ):

        summary = ["\nGALAXY LIGHT\n\n"]

        for radius in radii:
            luminosity = self.luminosity_within_circle_in_units(
                unit_luminosity=unit_luminosity,
                radius=radius,
                redshift_source=redshift_source,
                cosmology=cosmology,
                kwargs=kwargs,
            )

            summary += [
                text_util.within_radius_label_value_and_unit_string(
                    prefix=prefix + "luminosity",
                    radius=radius,
                    unit_length=unit_length,
                    value=luminosity,
                    unit_value=unit_luminosity,
                    whitespace=whitespace,
                )
            ]

        summary.append("\nLIGHT PROFILES:\n\n")

        for light_profile in self.light_profiles:
            summary += light_profile.summarize_in_units(
                radii=radii,
                whitespace=whitespace,
                unit_length=unit_length,
                unit_luminosity=unit_luminosity,
                redshift_profile=self.redshift,
                redshift_source=redshift_source,
                cosmology=cosmology,
                kwargs=kwargs,
            )

            summary += "\n"

        return summary

    def summarize_mass_profiles_in_units(
        self,
        radii,
        whitespace=80,
        prefix="",
        unit_length="arcsec",
        unit_mass="solMass",
        redshift_source=None,
        cosmology=cosmo.Planck15,
        **kwargs
    ):

        summary = ["\nGALAXY MASS\n\n"]

        einstein_radius = self.einstein_radius_in_units(
            unit_length=unit_length, cosmology=cosmology
        )

        summary += [
            af.text_util.label_value_and_unit_string(
                label=prefix + "einstein_radius",
                value=einstein_radius,
                unit=unit_length,
                whitespace=whitespace,
            )
        ]

        einstein_mass = self.einstein_mass_in_units(
            unit_mass=unit_mass,
            redshift_source=redshift_source,
            cosmology=cosmology,
            kwargs=kwargs,
        )

        summary += [
            af.text_util.label_value_and_unit_string(
                label=prefix + "einstein_mass",
                value=einstein_mass,
                unit=unit_mass,
                whitespace=whitespace,
            )
        ]

        for radius in radii:
            mass = self.mass_within_circle_in_units(
                unit_mass=unit_mass,
                radius=radius,
                redshift_source=redshift_source,
                cosmology=cosmology,
                kwargs=kwargs,
            )

            summary += [
                text_util.within_radius_label_value_and_unit_string(
                    prefix=prefix + "mass",
                    radius=radius,
                    unit_length=unit_length,
                    value=mass,
                    unit_value=unit_mass,
                    whitespace=whitespace,
                )
            ]

        summary += ["\nMASS PROFILES:\n\n"]

        for mass_profile in self.mass_profiles:
            summary += mass_profile.summarize_in_units(
                radii=radii,
                whitespace=whitespace,
                unit_length=unit_length,
                unit_mass=unit_mass,
                redshift_profile=self.redshift,
                redshift_source=redshift_source,
                cosmology=cosmology,
                kwargs=kwargs,
            )

            summary += "\n"

        return summary


class HyperGalaxy(object):
    _ids = count()

    def __init__(self, contribution_factor=0.0, noise_factor=0.0, noise_power=1.0):
        """ If a *Galaxy* is given a *HyperGalaxy* as an attribute, the noise-map in \
        the regions of the image that the galaxy is located will be hyper, to prevent \
        over-fitting of the galaxy.
        
        This is performed by first computing the hyper_galaxies-galaxy's 'contribution-map', \
        which determines the fraction of flux in every pixel of the image that can be \
        associated with this particular hyper_galaxies-galaxy. This is computed using \
        hyper_galaxies-hyper_galaxies set (e.g. fitting.fit_data.FitDataHyper), which includes  best-fit \
        unblurred_image_1d of the galaxy's light from a previous analysis phase.
         
        The *HyperGalaxy* class contains the hyper_galaxies-parameters which are associated \
        with this galaxy for scaling the noise-map.
        
        Parameters
        -----------
        contribution_factor : float
            Factor that adjusts how much of the galaxy's light is attributed to the
            contribution map.
        noise_factor : float
            Factor by which the noise-map is increased in the regions of the galaxy's
            contribution map.
        noise_power : float
            The power to which the contribution map is raised when scaling the
            noise-map.
        """
        self.contribution_factor = contribution_factor
        self.noise_factor = noise_factor
        self.noise_power = noise_power

        self.component_number = next(self._ids)

    def hyper_noise_map_from_hyper_images_and_noise_map(
        self, hyper_model_image, hyper_galaxy_image, noise_map
    ):
        contribution_map = self.contribution_map_from_hyper_images(
            hyper_model_image=hyper_model_image, hyper_galaxy_image=hyper_galaxy_image
        )
        return self.hyper_noise_map_from_contribution_map(
            noise_map=noise_map, contribution_map=contribution_map
        )

    def contribution_map_from_hyper_images(self, hyper_model_image, hyper_galaxy_image):
        """Compute the contribution map of a galaxy, which represents the fraction of
        flux in each pixel that the galaxy is attributed to contain, hyper to the
        *contribution_factor* hyper_galaxies-parameter.

        This is computed by dividing that galaxy's flux by the total flux in that \
        pixel and then scaling by the maximum flux such that the contribution map \
        ranges between 0 and 1.

        Parameters
        -----------
        hyper_model_image : ndarray
            The best-fit model image to the observed image from a previous analysis
            phase. This provides the total light attributed to each image pixel by the
            model.
        hyper_galaxy_image : ndarray
            A model image of the galaxy (from light profiles or an inversion) from a
            previous analysis phase.
        """
        contribution_map = np.divide(
            hyper_galaxy_image, np.add(hyper_model_image, self.contribution_factor)
        )
        contribution_map = np.divide(contribution_map, np.max(contribution_map))
        return contribution_map

    def hyper_noise_map_from_contribution_map(self, noise_map, contribution_map):
        """Compute a hyper galaxy hyper_galaxies noise-map from a baseline noise-map.

        This uses the galaxy contribution map and the *noise_factor* and *noise_power*
        hyper_galaxies-parameters.

        Parameters
        -----------
        noise_map : ndarray
            The observed noise-map (before scaling).
        contribution_map : ndarray
            The galaxy contribution map.
        """
        return self.noise_factor * (noise_map * contribution_map) ** self.noise_power

    def __eq__(self, other):
        if isinstance(other, HyperGalaxy):
            return (
                self.contribution_factor == other.contribution_factor
                and self.noise_factor == other.noise_factor
                and self.noise_power == other.noise_power
            )
        return False

    def __str__(self):
        return "\n".join(["{}: {}".format(k, v) for k, v in self.__dict__.items()])


class Redshift(float):
    def __new__(cls, redshift):
        # noinspection PyArgumentList
        return float.__new__(cls, redshift)

    def __init__(self, redshift):
        float.__init__(redshift)



import inspect
import os

from autolens import decorator_util
import numpy as np
from astropy.io import fits
from functools import wraps


class Memoizer(object):
    def __init__(self):
        """
        Class to store the results of a function given a set of inputs.
        """
        self.results = {}
        self.calls = 0
        self.arg_names = None

    def __call__(self, func):
        """
        Memoize decorator. Any time a function is called that a memoizer has been attached to its results are stored in
        the results dictionary or retrieved from the dictionary if the function has already been called with those
        arguments.

        Note that the same memoizer persists over all instances of a class. Any state for a given instance that is not
        given in the representation of that instance will be ignored. That is, it is possible that the memoizer will
        give incorrect results if instance state does not affect __str__ but does affect the value returned by the
        memoized method.

        Parameters
        ----------
        func: function
            A function for which results should be memoized

        Returns
        -------
        decorated : function
            A function that memoizes results
        """
        if self.arg_names is not None:
            raise AssertionError("Instantiate a new Memoizer for each function")
        self.arg_names = inspect.getfullargspec(func).args

        @wraps(func)
        def wrapper(*args, **kwargs):
            key = ", ".join(
                [
                    "('{}', {})".format(arg_name, arg)
                    for arg_name, arg in list(zip(self.arg_names, args))
                    + [(k, v) for k, v in kwargs.items()]
                ]
            )
            if key not in self.results:
                self.calls += 1
            self.results[key] = func(*args, **kwargs)
            return self.results[key]

        return wrapper


@decorator_util.jit()
def extracted_array_2d_from_array_2d_and_coordinates(array_2d, y0, y1, x0, x1):
    """Resize an array to a new size by extracting a sub-set of the array.

    The extracted input coordinates use NumPy convention, such that the upper values should be specified as +1 the \
    dimensions of the extracted array.

    In the example below, an array of size (5,5) is extracted using the coordinates y0=1, y1=4, x0=1, x1=4. This
    extracts an array of dimensions (3,3) and is equivalent to array_2d[1:4, 1:4].

    This function is necessary work with numba jit tags and is why a standard Numpy array extraction is not used.

    Parameters
    ----------
    array_2d : ndarray
        The 2D array that an array is extracted from.
    y0 : int
        The top row number (e.g. the higher y-coodinate) of the array that is extracted for the resize.
    y1 : int
        The bottom row number (e.g. the lower y-coodinate) of the array that is extracted for the resize.
    x0 : int
        The left column number (e.g. the lower x-coodinate) of the array that is extracted for the resize.
    x1 : int
        The right column number (e.g. the higher x-coodinate) of the array that is extracted for the resize.

    Returns
    -------
    ndarray
        The extracted 2D array from the input 2D array.

    Examples
    --------
    array_2d = np.ones((5,5))
    extracted_array = extract_array_2d(array_2d=array_2d, y0=1, y1=4, x0=1, x1=4)
    """

    new_shape = (y1 - y0, x1 - x0)

    resized_array = np.zeros(shape=new_shape)

    for y_resized, y in enumerate(range(y0, y1)):
        for x_resized, x in enumerate(range(x0, x1)):
            if (
                y >= 0
                and x >= 0
                and y <= array_2d.shape[0] - 1
                and x <= array_2d.shape[1] - 1
            ):
                resized_array[y_resized, x_resized] = array_2d[y, x]

    return resized_array


@decorator_util.jit()
def resized_array_2d_from_array_2d_and_resized_shape(
    array_2d, resized_shape, origin=(-1, -1), pad_value=0.0
):
    """Resize an array to a new size around a central pixel.

    If the origin (e.g. the central pixel) of the resized array is not specified, the central pixel of the array is \
    calculated automatically. For example, a (5,5) array's central pixel is (2,2). For even dimensions the central \
    pixel is assumed to be the lower indexed value, e.g. a (6,4) array's central pixel is calculated as (2,1).

    The default origin is (-1, -1) because numba requires that the function input is the same type throughout the \
    function, thus a default 'None' value cannot be used.

    Parameters
    ----------
    array_2d : ndarray
        The 2D array that is resized.
    resized_shape : (int, int)
        The (y,x) new pixel dimension of the trimmed array.
    origin : (int, int)
        The oigin of the resized array, e.g. the central pixel around which the array is extracted.
    pad_value : float
        If the reszied array is bigger in size than the input array, the value the padded edge values are filled in \
        using.

    Returns
    -------
    ndarray
        The resized 2D array from the input 2D array.

    Examples
    --------
    array_2d = np.ones((5,5))
    resize_array = resize_array_2d(array_2d=array_2d, new_shape=(2,2), origin=(2, 2))
    """

    y_is_even = int(array_2d.shape[0]) % 2 == 0
    x_is_even = int(array_2d.shape[1]) % 2 == 0

    if origin == (-1, -1):

        if y_is_even:
            y_centre = int(array_2d.shape[0] / 2)
        elif not y_is_even:
            y_centre = int(array_2d.shape[0] / 2)

        if x_is_even:
            x_centre = int(array_2d.shape[1] / 2)
        elif not x_is_even:
            x_centre = int(array_2d.shape[1] / 2)

        origin = (y_centre, x_centre)

    resized_array = np.zeros(shape=resized_shape)

    if y_is_even:
        y_min = origin[0] - int(resized_shape[0] / 2)
        y_max = origin[0] + int((resized_shape[0] / 2)) + 1
    elif not y_is_even:
        y_min = origin[0] - int(resized_shape[0] / 2)
        y_max = origin[0] + int((resized_shape[0] / 2)) + 1

    if x_is_even:
        x_min = origin[1] - int(resized_shape[1] / 2)
        x_max = origin[1] + int((resized_shape[1] / 2)) + 1
    elif not x_is_even:
        x_min = origin[1] - int(resized_shape[1] / 2)
        x_max = origin[1] + int((resized_shape[1] / 2)) + 1

    for y_resized, y in enumerate(range(y_min, y_max)):
        for x_resized, x in enumerate(range(x_min, x_max)):
            if y >= 0 and y < array_2d.shape[0] and x >= 0 and x < array_2d.shape[1]:
                if (
                    y_resized >= 0
                    and y_resized < resized_shape[0]
                    and x_resized >= 0
                    and x_resized < resized_shape[1]
                ):
                    resized_array[y_resized, x_resized] = array_2d[y, x]
            else:
                if (
                    y_resized >= 0
                    and y_resized < resized_shape[0]
                    and x_resized >= 0
                    and x_resized < resized_shape[1]
                ):
                    resized_array[y_resized, x_resized] = pad_value

    return resized_array


@decorator_util.jit()
def replace_noise_map_2d_values_where_image_2d_values_are_negative(
    image_2d, noise_map_2d, target_signal_to_noise=2.0
):
    """If the values of a 2D image array are negative, this function replaces the corresponding 2D noise-map array \
    values to meet a specified target to noise value.

    This routine is necessary because of anomolous values in images which come from our HST ACS data_type-reduction \
    pipeline, where image-pixels with negative values (e.g. due to the background sky subtraction) have extremely \
    small noise values, which inflate their signal-to-noise values and chi-squared contributions in the modeling.

    Parameters
    ----------
    image_2d : ndarray
        The 2D image array used to locate the pixel indexes in the noise-map which are replaced.
    noise_map_2d : ndarray
        The 2D noise-map array whose values are replaced.
    target_signal_to_noise : float
        The target signal-to-noise the noise-map valueus are changed to.

    Returns
    -------
    ndarray
        The 2D noise-map with values changed.

    Examples
    --------
    image_2d = np.ones((5,5))
    image_2d[2,2] = -1.0
    noise_map_2d = np.ones((5,5))

    noise_map_2d_replaced = replace_noise_map_2d_values_where_image_2d_values_are_negative(
        image_2d=image_2d, noise_map_2d=noise_map_2d, target_signal_to_noise=2.0):
    """
    for y in range(image_2d.shape[0]):
        for x in range(image_2d.shape[1]):
            if image_2d[y, x] < 0.0:
                absolute_signal_to_noise = np.abs(image_2d[y, x]) / noise_map_2d[y, x]
                if absolute_signal_to_noise >= target_signal_to_noise:
                    noise_map_2d[y, x] = np.abs(image_2d[y, x]) / target_signal_to_noise

    return noise_map_2d


def numpy_array_1d_to_fits(array_1d, file_path, overwrite=False):
    """Write a 1D NumPy array to a .fits file.

    Before outputting a NumPy array, the array is flipped upside-down using np.flipud. This is so that the arrays \
    appear the same orientation as .fits files loaded in DS9.

    Parameters
    ----------
    array_2d : ndarray
        The 2D array that is written to fits.
    file_path : str
        The full path of the file that is output, including the file name and '.fits' extension.
    overwrite : bool
        If True and a file already exists with the input file_path the .fits file is overwritten. If False, an error \
        will be raised.

    Returns
    -------
    None

    Examples
    --------
    array_2d = np.ones((5,5))
    numpy_array_to_fits(array_2d=array_2d, file_path='/path/to/file/filename.fits', overwrite=True)
    """
    if overwrite and os.path.exists(file_path):
        os.remove(file_path)

    new_hdr = fits.Header()
    hdu = fits.PrimaryHDU(array_1d, new_hdr)
    hdu.writeto(file_path)


def numpy_array_1d_from_fits(file_path, hdu):
    """Read a 2D NumPy array to a .fits file.

    After loading the NumPy array, the array is flipped upside-down using np.flipud. This is so that the arrays \
    appear the same orientation as .fits files loaded in DS9.

    Parameters
    ----------
    file_path : str
        The full path of the file that is loaded, including the file name and '.fits' extension.
    hdu : int
        The HDU extension of the array that is loaded from the .fits file.

    Returns
    -------
    ndarray
        The NumPy array that is loaded from the .fits file.

    Examples
    --------
    array_2d = numpy_array_from_fits(file_path='/path/to/file/filename.fits', hdu=0)
    """
    hdu_list = fits.open(file_path)
    return np.array(hdu_list[hdu].data)


def numpy_array_2d_to_fits(array_2d, file_path, overwrite=False):
    """Write a 2D NumPy array to a .fits file.

    Before outputting a NumPy array, the array is flipped upside-down using np.flipud. This is so that the arrays \
    appear the same orientation as .fits files loaded in DS9.

    Parameters
    ----------
    array_2d : ndarray
        The 2D array that is written to fits.
    file_path : str
        The full path of the file that is output, including the file name and '.fits' extension.
    overwrite : bool
        If True and a file already exists with the input file_path the .fits file is overwritten. If False, an error \
        will be raised.

    Returns
    -------
    None

    Examples
    --------
    array_2d = np.ones((5,5))
    numpy_array_to_fits(array_2d=array_2d, file_path='/path/to/file/filename.fits', overwrite=True)
    """
    if overwrite and os.path.exists(file_path):
        os.remove(file_path)

    new_hdr = fits.Header()
    hdu = fits.PrimaryHDU(np.flipud(array_2d), new_hdr)
    hdu.writeto(file_path)


def numpy_array_2d_from_fits(file_path, hdu):
    """Read a 2D NumPy array to a .fits file.

    After loading the NumPy array, the array is flipped upside-down using np.flipud. This is so that the arrays \
    appear the same orientation as .fits files loaded in DS9.

    Parameters
    ----------
    file_path : str
        The full path of the file that is loaded, including the file name and '.fits' extension.
    hdu : int
        The HDU extension of the array that is loaded from the .fits file.

    Returns
    -------
    ndarray
        The NumPy array that is loaded from the .fits file.

    Examples
    --------
    array_2d = numpy_array_from_fits(file_path='/path/to/file/filename.fits', hdu=0)
    """
    hdu_list = fits.open(file_path)
    return np.flipud(np.array(hdu_list[hdu].data))
